/*
 * This file implements HRW redirection scheme for CoDNS
 * 
 */
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "queue.h"
#include "codns_int.h"
#include "debug.h"
#include "codnslog.h"

#define VALID_DIST(x) ((x)->pi_avgLocalLookupMS != INT_MAX && \
                       (x)->pi_avgPingTime != INT_MAX)
#define DISTANCE(x)   ((x)->pi_avgLocalLookupMS + (x)->pi_avgPingTime)

/* sorted list of neighbor nodes */
static TAILQ_HEAD(, PeerInfo) 
       allNeighbors = LIST_HEAD_INITIALIZER(allNeighbors);

/*-----------------------------------------------------------------*/
inline int 
Closer(const PeerInfo* node, const PeerInfo* walk)
{
  if (!VALID_DIST(walk) || DISTANCE(node) <= DISTANCE(walk))
    return 1;
  return 0;
}
/*-----------------------------------------------------------------*/
inline int
SameServer(int i, int j)
{
#define SAME_ADDR(x, y)   ((x).s_addr == (y).s_addr)
#define SAME_SERVER(x, y) ((x).s_addr != 0 &&                         \
                           (SAME_ADDR((x), (y)->pi_nameServers[0]) || \
			    SAME_ADDR((x), (y)->pi_nameServers[1])))

 const PeerInfo *x = &g_allPeers[i]; 
 const PeerInfo *y = &g_allPeers[j]; 

 if (SAME_SERVER(x->pi_nameServers[0], y) ||
     SAME_SERVER(x->pi_nameServers[1], y))
   return 1;

 return 0;
}
/*-----------------------------------------------------------------*/
static int
IsPeerAlive(const PeerInfo* pi)
{
  /* see if this node is alive */
#define CUTOFF_SECS 2
  int ackHistory =  pi->pi_hbGoodAckHistory & hbAckHistoryMask;
  time_t waitTime;

  /* if we've got all ACKs, then it's alive */
  if (ackHistory == hbAckHistoryMask)
    return TRUE;

  /* if we've got all ACKs except for the last one, it's alive */
  waitTime = (timeex(NULL) - pi->pi_hbLastSentTime);
  if ((ackHistory | 1) == hbAckHistoryMask &&
      waitTime < CUTOFF_SECS)
    return TRUE;

  /* if we've seen the latest ACK, it's alive */
  if ((ackHistory & 1) && waitTime < (10*CUTOFF_SECS))
    return TRUE;

  return FALSE;
}
/*-----------------------------------------------------------------*/
static void
DoEarlyRecovery(void)
{
  /* we observed that temporarily all netwok packets get lost in some
     nodes. To recover from this kind of disaster quickly, we restore
     the ACK history here */
 int i;
 PeerInfo* pi;
#define UPPER_BIT_MASK (0xffff0000)

/* we reset the history */
 for (i = 0; i < g_numPeers; i++) {
   pi = &g_allPeers[i];

   /* we only restore temporary problematic nodes */
   if ((pi->pi_hbGoodAckHistory & UPPER_BIT_MASK) == UPPER_BIT_MASK)
     pi->pi_hbGoodAckHistory |= (hbAckHistoryMask & (hbAckHistoryMask-1));
 }
}
/*-----------------------------------------------------------------*/
void
InitNeighborsList(void)
{
  PeerInfo* node;
  int i;

  TAILQ_INIT(&allNeighbors);
  for (i = 0; i < g_numPeers; i++) {
    node = &g_allPeers[i];
    TAILQ_INSERT_HEAD(&allNeighbors, node, node_all);
  }
}
/*-----------------------------------------------------------------*/
void 
UpdateNeighborsList(PeerInfo* node) 
{
  PeerInfo* walk;

  /* first, remove this node from the list */
  TAILQ_REMOVE(&allNeighbors, node, node_all);

  walk = TAILQ_FIRST(&allNeighbors);
  if (Closer(node, walk)) {
    TAILQ_INSERT_HEAD(&allNeighbors, node, node_all);
    return;
  }

  /* walk the list until we're in sorted order */
  TAILQ_FOREACH(walk, &allNeighbors, node_all) {
    if (Closer(node, walk)) {
     TAILQ_INSERT_BEFORE(walk, node, node_all);
     return;
    }
    if (TAILQ_NEXT(walk, node_all) == NULL) {
      TAILQ_INSERT_TAIL(&allNeighbors, node, node_all);
      return;
    }
  }

  TRACE_L1("walked past end\n");
  exit(-1);
}
/*-----------------------------------------------------------------*/
void
GetNeighborsSnapshot(int *array, int *maxNum, int all)
{
  PeerInfo* walk;
  int cnt = 0, i, duplicate = 0;

  TAILQ_FOREACH(walk, &allNeighbors, node_all) {
    if (!all && !VALID_DIST(walk))
      break;

    if (!all && !IsPeerAlive(walk)) 
      continue;
      
    /* it should not be the one with same server sets
       of the nodes that we've added */
    if (!all) {
      for (duplicate = 0, i = 0; i < cnt; i++) {
	if (SameServer(array[i], walk->pi_idx)) {
	  duplicate = 1;
	  break;
	}
      }
    }

    if (!duplicate) {
      array[cnt++] = walk->pi_idx;
      if (!all && cnt == MAX_NEIGHBORS) 
	break;
    }
  }	
  *maxNum = cnt;
} 
/*-----------------------------------------------------------------*/
int 
PR_HRWPickServer(char *in_name)
{
  /* Proximity-aware R-HRW
     select a peer node for remote DNS queries
     this uses HRW on the group of near nodes */

  static int totReplica = -1;
  unsigned int urlHash;
  int j, rand_num, svr_no = -1, ready_servers = 0;
  int total_rep;
  PeerInfo* walk;
  ServerWeight server_weights[MAX_PEERS];
  int i, duplicate;

  if (totReplica == -1) 
    totReplica = GetOptReplicaNum(g_numPeers);

  urlHash = HashString(in_name, 0, TRUE);
  TAILQ_FOREACH(walk, &allNeighbors, node_all) {

    /* don't have search any more ? */
    if (!VALID_DIST(walk) || ready_servers >= MAX_NEIGHBORS)
      break;
    
    /* it should be alive and it not be myself */
    if (!IsPeerAlive(walk) || (walk->pi_idx == g_localHostIdx)) 
      continue;

    /* also, it should not be the one with same server sets
       of the nodes that we've added */
    for (duplicate = 0, i = 0; i < ready_servers; i++) {
      if (SameServer(server_weights[i].index, walk->pi_idx)) {
	duplicate = 1;
	break;
      }
    }

    /* now, record the valid candidate */
    if (!duplicate) {
      server_weights[ready_servers].index = walk->pi_idx;
      server_weights[ready_servers].weight = 
	WrandHash(urlHash, walk->pi_addrHash);
      ready_servers++;
    }
  }

  /* all nodes are currenly dead ? */
  if (walk == NULL && ready_servers == 0) 
    DoEarlyRecovery();

  /* in case no servers are up, we'll return -1 */
  if (ready_servers > 0) {
    quickSort(server_weights, 0, ready_servers -1);
    total_rep = __min(ready_servers, totReplica);

    /* randomly select a server from totReplica servers */
    rand_num = random() & 0xffffffff;
    j = rand_num % total_rep;
    svr_no = server_weights[j].index;
  }

  /* if the server is not up, simply let redirector contact origin server? */
  return svr_no;
}
/*-----------------------------------------------------------------*/
