#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef OS_DARWIN
#include <nameser.h>
#endif
#include <resolv.h>
#include "queue.h"
#include "applib.h"
#include "codns_int.h"
#include "codnshbt.h"
#include "HRW_codns.h"
#include "codnsstat.h"
#include "debug.h"
#include "codnslog.h"
#include "ports.h"

/* version history
    1 : start version (08/xx/03 ?)
    2 : from 12/30/03 ~ 10/27/04
    3 : supports TTL of DNS record
    4 : supports multiple A records
    5 : communicate in network byte order
*/
#define CODNS_VERS 5		  

/* global variables */
PeerInfo g_allPeers[MAX_PEERS];   /* all peers' information */
int g_numPeers = 0;               /* number of all peers */
int g_localHostIdx;               /* local host name idx at g_allPeers[] */
int g_avgLocalLookupMS;           /* average local name lookup time */
struct timeval g_currentTime;
int g_maxNeighbors = 0;

/* my host name */
#define MAX_HOSTNAME 1024
char g_localHostname[MAX_HOSTNAME]; 

/* fds for logging processes */
int g_logFd;                      /* basic debug trace */
HANDLE hdebugLog;		  /* for debugging log */

/* persistent connection file descriptor list */
typedef struct PersistConnFD {
  int pc_fd;                         /* persistent connection socket fd */
  LIST_ENTRY(PersistConnFD) pc_all;  /* for maintaining the FD list */
  LIST_ENTRY(PersistConnFD) pc_dead; /* for collecting dead entries */
}PersistConnFD;

static TAILQ_HEAD(, QuerySource) 
       allQueries = TAILQ_HEAD_INITIALIZER(allQueries);
static LIST_HEAD(WaitList, QuerySource) 
       udpWait = LIST_HEAD_INITIALIZER(udpWait),
       tcpWait = LIST_HEAD_INITIALIZER(tcpWait);
static LIST_HEAD(PersistConnList, PersistConnFD) 
       allFDs = LIST_HEAD_INITIALIZER(allFDs);

#define MIN_NUM_SLAVES 6
#define MAX_NUM_SLAVES 20
typedef struct SlaveInfo {
  int si_sock;			/* slave's socket */
  int si_inUse;
  QuerySource *si_qs;		/* where to respond */
  LIST_ENTRY(SlaveInfo) si_free;
} SlaveInfo;

#define MAX_UDP_WAITSIZE 100	/* discard remote queries after this many */
static int g_udpWaitSize;
static int g_slaveInUse;

static SlaveInfo g_slave[MAX_NUM_SLAVES];
static LIST_HEAD(, SlaveInfo) 
       freeSlaves = LIST_HEAD_INITIALIZER(freeSlaves);

typedef void SlaveFunc(int fd, void *token);

static struct timeval g_nextGlobalTimeout;

/* waiting period(secs) between hearbeats */
static int g_HeartbeatTimeout = 1; 

/* number of slaves */
static int g_NumSlaves = MAX_NUM_SLAVES;

/* keep track of how often udp won in recent queries */
static int g_udpWinsHistory;

/* initial timeout for remote queries in usecs */
static int g_udpTimeout = 200000;      /* set to 200 ms as a default */

/* set current version of configuration file 
   current format is "%04d%02d%02d%2d%2d", year, month, day, hour, min
   so, we need 64 bits ex) 200502240712 */
static long long g_curConfigVersion = 0;

/* configuration file auto update URL */
char* g_autoUpdateHost = NULL;
char* g_autoUpdateFile = NULL;

/* for passing the peer lookup result */
typedef struct PeerLookupResult {
  long long      plr_configVer;   /* configuration file version */
  int            plr_idx;         /* index of this peer node */
  struct in_addr plr_address;     /* ip address */
} PeerLookupResult;

/* base directory */
char g_basePath[MAX_PATH];
#define PREPEND_BASEPATH(x, relpath) \
   snprintf(x, sizeof(x), "%s/%s", g_basePath, relpath);

#ifndef FREE_MEM
#define FREE_MEM(x) if (x) { free(x); x = NULL; }
#endif

/* defined in updateconfig.c */
extern void CreateConfigFileAutoUpdater(void);

/*-----------------------------------------------------------------*/
static int
AddPersistConn(int fd)
{
  PersistConnFD* pc;
  
  if ((pc = calloc(1, sizeof(PersistConnFD))) != NULL) {
    pc->pc_fd = fd;
    LIST_INSERT_HEAD(&allFDs, pc, pc_all);
    return 0;
  }
  return -1;
}
/*-----------------------------------------------------------------*/
static int
RemovePersistConn(int fd) 
{
  PersistConnFD* walk;
  LIST_FOREACH(walk, &allFDs, pc_all) {
    if (walk->pc_fd == fd) {
      TRACE_L("found an entry fd = %d\n", fd);
      break;
    }
  }
  if (walk) {
    LIST_REMOVE(walk, pc_all);
    FREE_MEM(walk);
    return 0;
  }
  return -1;
}
/*-----------------------------------------------------------------*/
static QuerySource *
ReadQuery(int fd, int isUDP, int* need_close)
{
  /* takes either a UDP or local socket, and returns a QuerySource
     structure for it. Does NOT add it to any queues since a UDP
     message may be a response rather than a query. For local queries,
     generates the random token itself */
  int res = 0;
  QuerySource *qs = calloc(1, sizeof(QuerySource));
  int fromLen = sizeof(struct sockaddr_in);

  if (!qs) {
    TRACE_LX1( "out of mem reading query\n");
    NiceExit(-1, "out of mem reading query");
  }

  qs->qs_fd = fd;
  qs->qs_isUDP = isUDP;
  if (need_close)
    *need_close = FALSE;

  if (isUDP) {
    res = recvfrom(fd, &qs->qs_query, sizeof(NSQuery), 0,
		   (struct sockaddr *) &qs->qs_origin, (socklen_t *)&fromLen);
    if (res < 0) {
      if (errno == EAGAIN) {
	TRACE_LX1("Data doesn't seem to be available now\n");
      } else {
	TRACE_LX1("Some weird error happened\n");
      }
    } else {
      ntoh_NSQuery(&qs->qs_query);
      if (qs->qs_query.nsq_version != CODNS_VERS) {
	TRACE_LX1("version mismatch is detected\n");
	res = 0;		/* ignore anything of a different version */
      }
    }
  }
  else {
    /* read the query info first,
       then read the domain name string */
    int sig;

    /* read the signature byte first */
    if ((res = read(fd, &sig, sizeof(sig))) == sizeof(sig)) {

      if (sig != SIG_SPLIT_TRANSACTION) { 
	/* plain-text query */
	char *name = qs->qs_query.nsq_name;

	memcpy(name, &sig, sizeof(sig));
	name += sizeof(sig);
	if ((res = read(fd, name, 
			sizeof(qs->qs_query.nsq_name) - sizeof(sig))) > 0)
	  res += sizeof(sig); 
      } else {        
	/* split-transaction query */
	int size = sizeof(LocalQueryInfo);

	if ((res = read(fd, &qs->qs_queryinfo, size)) == size) {
	  ntoh_LocalQueryInfo(&qs->qs_queryinfo);

	  size = qs->qs_queryinfo.lqi_size;
	  if (size > MAX_NSQUERY_NAME) {
	    TRACE_LX("name length too big, len %d\n", size);
	    res = 0;  /* force it to close */
	  } else if ((res = read(fd, qs->qs_query.nsq_name, size)) != size 
		     && (res != 0)) {
	    TRACE_LX("name length differs : read %d, len %d\n", res, size);
	    res = 0; /* force it to close */
	  }
	} else if (res != 0) {
	  TRACE_LX("query size differs : read %d, len %d\n", res, size);
	}
      }

      /* clean up the string for both cases */
      if (res >= 0) /* terminate it just in case local didn't */
	qs->qs_query.nsq_name[res] = 0;

      while (res > 0 && isspace((int)qs->qs_query.nsq_name[res-1])) {
	qs->qs_query.nsq_name[res-1] = 0;
	res--;
      }
      qs->qs_query.nsq_randomToken = random();
      qs->qs_query.nsq_version = CODNS_VERS;
    }

    /* connetion needs to be closed ? */
    if (res == 0 && need_close) 
      *need_close = TRUE;
  }
  
  /* if we've got an error or connection-close, 
     then we return NULL */
  if (res < 1) {
    if (res < 0)
      TRACE_LX( "had res %d in ReadQuery\n", res);
    FREE_MEM(qs);
    return(NULL);
  }

  /* terminate it for safety */
  qs->qs_query.nsq_name[MAX_NSQUERY_NAME-1] = 0;
  return(qs);
}
/*-----------------------------------------------------------------*/
static QuerySource *
QueryMatch(QuerySource *test, int exactMatch, int udpOK)
{
  /* given a query, finds a matching query among all outstanding
     queries. Used to see if an incoming response corresponds to an
     extant waiting query */
  QuerySource *walk;
#define TOKEN_MATCH(x, y) ((x)->qs_query.nsq_randomToken == \
                           (y)->qs_query.nsq_randomToken)

  TAILQ_FOREACH(walk, &allQueries, qs_all) {
    if ((udpOK || !walk->qs_isUDP) &&   
	(!exactMatch || TOKEN_MATCH(test, walk)) &&
	(strcmp(test->qs_query.nsq_name, walk->qs_query.nsq_name) == 0))
      return(walk);
  }
  return(NULL);
}
/*-----------------------------------------------------------------*/
static int
SendResponse(QuerySource *qs, fd_set* pmasterSet, int *need_close)
{
  /* sends a response back to the client, using the appropriate
     calling mechanism. returns 1 if response was already sent 
     
     In the case of one-time connection, when masterSet is specified,
     this function tries to close it and remove fd from the master fd set.
     need_close will be set to TRUE, when it needs to be closed.
  */

  if (qs->qs_fd < 0) 
    return(1);			/* already responded */

  if (qs->qs_isUDP) {
    /* only send a UDP answer if we were able to really resolve
       it. otherwise, sending a fast but wrong answer could cause
       others to experience problems */
    if (qs->qs_query.nsq_result.nsq_address[0].s_addr != 0) {
      int res, size;
      
      size = sizeof(NSQuery);
      hton_NSQuery(&qs->qs_query);
      if ((res = sendto(qs->qs_fd, &qs->qs_query, size, 0,
			(struct sockaddr *) &qs->qs_origin,
			sizeof(struct sockaddr_in))) != size) {
	TRACE_LX("weird sendto size - %d, %d, %d\n", res, size, errno);
      }
      ntoh_NSQuery(&qs->qs_query);
    }
  }
  else {
    /* the client just wants the IP address in readable form */
    LocalQueryResult result;

    if (qs->qs_queryinfo.lqi_size > 0) { /* split-transaction */
      result.lq_id = qs->qs_queryinfo.lqi_id;
      result.lq_ttl = qs->qs_query.nsq_result.nsq_ttl;
      memcpy(result.lq_address, qs->qs_query.nsq_result.nsq_address, 
	     MAX_ANSWERS * sizeof(struct in_addr));
      
      hton_LocalQueryResult(&result);
      if (write(qs->qs_fd, &result, sizeof(result)) != sizeof(result))
	TRACE_LX1( "write had weird len\n");
    } else {                            /* plain text query */
      char respBuf[512];
      int len;

      /* make a reply in a human-readable form */
      len = snprintf(respBuf, sizeof(respBuf), "%s\n",
		     inet_ntoa(qs->qs_query.nsq_result.nsq_address[0]));
      if (write(qs->qs_fd, respBuf, len) != len)
	TRACE_LX1( "write had weird len\n");

      /* we're done with reply, 
	 so we need to close the connection */
      if (pmasterSet) {
	FD_CLR(qs->qs_fd, pmasterSet);
	close(qs->qs_fd);
	RemovePersistConn(qs->qs_fd);
      } else if (need_close) {
	*need_close = TRUE;
      }
    }
  }

  qs->qs_fd = -1;		/* mark it as answered */
  return(0);
}
/*-----------------------------------------------------------------*/
static int
SendAllResponse(QuerySource *qs, fd_set* pmasterSet) 
{
  /* preserve the same semantics of SendResponse(),
     but this one tries to satisfy my family list too */

  QuerySource* walk;

  /* send all response for my family */
  LIST_FOREACH(walk, &qs->qs_familyList, qs_family) {
    walk->qs_query.nsq_result = qs->qs_query.nsq_result;
    SendResponse(walk, pmasterSet, NULL);
  }

  /* clean up the sent entries */
  while ((walk = LIST_FIRST(&qs->qs_familyList)) != NULL) {
    LIST_REMOVE(walk, qs_family);
    FREE_MEM(walk);
  }
  return SendResponse(qs, pmasterSet, NULL);
}
/*-----------------------------------------------------------------*/
static void
FreeQuery(QuerySource *qs)
{
  /* frees the query and removes it from all lists */
  /* TRACE_L( "entering %s\n", __FUNCTION__); */
  static int numFreed;
  static int numWithUDPs;
  static int numUDPsTotal;
  static int numWithSlave;
  static int numSlaveLost;
  static time_t lastStartTime;

  if (lastStartTime/60 != g_currentTime.tv_sec/60) {
    if (numFreed) {
      TRACE_L("%d %s", (int) lastStartTime, ctime(&lastStartTime));
      TRACE_L("%d freed, %d w/slave, %d g_slave lost, %d w/UDP, %d UDPs\n",
	    numFreed, numWithSlave, numSlaveLost, numWithUDPs, numUDPsTotal);
    }
    numFreed = numWithUDPs = numUDPsTotal = numWithSlave = numSlaveLost = 0;
    lastStartTime = g_currentTime.tv_sec;
  }

  if (!qs->qs_isUDP) {
    numFreed++;
    if (qs->qs_numUDPs)
      numWithUDPs++;
    numUDPsTotal += qs->qs_numUDPs;
    numWithSlave += qs->qs_hitSlave;
    numSlaveLost += qs->qs_slaveLost;
  }

  if (qs->qs_onWaiting)
    LIST_REMOVE(qs, qs_waiting);
  TAILQ_REMOVE(&allQueries, qs, qs_all);
  FREE_MEM(qs);
}
/*-----------------------------------------------------------------*/
static void
FreeSlave(int whichSlave)
{
  /* disposes of the query, marks the g_slave as free, puts it back
     on the free list */
  SlaveInfo *si = &g_slave[whichSlave];

  /* TRACE_L( "entering %s\n", __FUNCTION__); */
  if (!si->si_inUse) {
    TRACE_LX1( "trying to free a free g_slave\n");
    NiceExit(-1, "trying to free a free g_slave");
  }

  si->si_inUse = 0;
  FreeQuery(si->si_qs);
  LIST_INSERT_HEAD(&freeSlaves, si, si_free);
  g_slaveInUse--;
}
/*-----------------------------------------------------------------*/
static void
LogSlaveMain(int fd, void *token)
{
  /* reads in log traces from CoDNS main process and write into a file
     designed to avoid blocking in the main process */

  char buffer[2048];
  int res;
  time_t flushTime =  timeex(NULL) + 5;
  
  while (1) {

    res = read(fd, buffer, sizeof(buffer));
    if (res <= 0) {
      TRACE("read failed, res = %d\n", res);
      NiceExit(-1, "read failed");
    }
    WriteLog(hdebugLog, buffer, res, FALSE);

    if (timeex(NULL) >= flushTime) {
    /* check if we have to wrap up yesterday's log and reopen today's 
       we don't have to be too exact, so we check this every five secodns */
      DailyReopenLogF(hdebugLog);

      /* flush logs */
      FlushLogF(hdebugLog);
      flushTime = timeex(NULL) + 5;
    }
  }
}
/*-----------------------------------------------------------------*/
static void 
RotateServers(void)
{
  /* rotate the search order of the servers listed 
     in /etc/resolv.conf */
  struct sockaddr_in temp;
  int i;

  if (_res.nscount == 1)
    return;

  temp = _res.nsaddr_list[0];
  for (i = 0; i < _res.nscount - 1; i++) 
    _res.nsaddr_list[i] = _res.nsaddr_list[i+1];

  _res.nsaddr_list[i] = temp;
}
/*-----------------------------------------------------------------*/
static void
IgnoreLoopBackServer(void)
{ 
  struct sockaddr_in temp;
  int i,j ;

  for(i = 0; i < _res.nscount; i++) {
    if (_res.nsaddr_list[i].sin_addr.s_addr == htonl(INADDR_LOOPBACK)) {
      //TRACEX("found a loopback server\n");
      break;
    }
  }
  if (i == _res.nscount) /* not found */
    return;

  /* place the loopback interface in the last slot */
  _res.nscount--;
  if (i < _res.nscount) {
    temp = _res.nsaddr_list[i];
    for (j = i; j < _res.nscount; j++)
      _res.nsaddr_list[j] = _res.nsaddr_list[j+1];
    _res.nsaddr_list[j] = temp;
  }
}
/*-----------------------------------------------------------------*/
static void
DNSSlaveMain(int fd, void *token)
{
  /* very self-contained - reads request on its socket, resolves it,
     returns response */
  char name[MAX_NSQUERY_NAME];
  struct timeval start, end;
#define FAILURE_SEC 5000

  res_init();
  IgnoreLoopBackServer();

  while (1) {
    int i;
    int res;
    int dnsWorked = TRUE;
    NSQueryResult result;
    struct hostent *tempHost;

    if ((res = read(fd, name, sizeof(name))) < 0) {
      TRACE_LX("weirdness in reading name, fd = %d, res = %d\n", fd, res);
      NiceExit(-1, "weirdness in reading name");
    }
    if (res == 0)
      exit(0);

    /* measure the local server latency 
       we think it's a failure when the latency is over 5 secs */
    gettimeofdayex(&start, NULL);
    tempHost = gethostbynameX(name);
    gettimeofdayex(&end, NULL);

    if (DiffTimesToMSec(end, start) >= FAILURE_SEC)
      dnsWorked = FALSE;

    memset(&result, 0, sizeof(result));
    if (tempHost != NULL) {
      result.nsq_ttl = (_dns_ttl_ > -1) ? _dns_ttl_ : 0;
      
      for (i = 0; i < MAX_ANSWERS && tempHost->h_addr_list[i]; i++) {
	result.nsq_address[i] = 
	  *((struct in_addr *)(tempHost->h_addr_list[i]));
      }

    } else {  
      /* wait one second for remote response */
      sleep(1);  
    }

    /* send our response */
    write(fd, &result, sizeof(result));

    /* if dns failed, we rotate the local server list */
    if (dnsWorked == FALSE) 
      RotateServers();
  }
}
/*-----------------------------------------------------------------*/
static int
CreateSlave(SlaveFunc *func, void *token, int doTrace)
{
  /* creates a pipe, forks off a new process, closes the appropriate
     ends, and returns the parent's fd back to the caller. The child
     then waits for requests and processes them */
  int socks[2];
  int pid;

  if (socketpair(AF_UNIX, SOCK_STREAM, 0, socks) < 0) {
    TRACE_LX1("failed on socketpair\n");
    NiceExit(-1, "failed on socketpair");
  }

  if ((pid = fork()) < 0) {
    TRACE_LX1("failed on socketpair\n");
    NiceExit(-1, "failed on socketpair");
  }
  if (pid == 0) {
    /* we're in child */
    close(socks[0]);
    func(socks[1], token);
    /* we should never get here */
    TRACE_LX1( "slave main returned\n");
    NiceExit(-1, "slave main returned");
  }
  if (doTrace)
    TRACE_L("created slave %d\n", pid);
  close(socks[1]);
  return(socks[0]);
}
/*-----------------------------------------------------------------*/
/* configuration file option fields */
#define OPT_TIMEOUT    "timeout"
#define OPT_VERSION    "version"
#define OPT_NODLIST    "nodelist"
#define OPT_AUTOUPDATE "autoupdate"
#define OPT_HB_TIMEOUT "hbtimeout"
#define OPT_NUM_SLAVES "numslaves"
int
IsConfigFileUpdated(const char* filename)
{
  /* see if the configuration file is updated */
  char confFile[MAX_PATH];
  char line[1024];
  char *temp;
  FILE* list = NULL;
  long long ver = 0;
  struct stat buf;
  static time_t lastModTime = 0;

  if (filename == NULL) {
    PREPEND_BASEPATH(confFile, NEWCONFIGFILE);
  } else {
    snprintf(confFile, sizeof(confFile), filename);
  }

  /* get the file stat */
  if (stat(confFile, &buf) < 0) 
    return FALSE;

  /* see if it's readable */
  if ((buf.st_mode & S_IRUSR) == 0) {
    TRACE_LX("%s file is not readable! "
	    "please add the read privilege\n", confFile);
    return FALSE;
  }
  
  /* check last modified time has changed */
  if (filename == NULL) {
    /* if it's regular checkup, we compare last modified time */
    if (lastModTime == buf.st_mtime) 
      return FALSE;
    lastModTime = buf.st_mtime;
  }
  
  if ((list = fopen(confFile, "r")) == NULL) {
    TRACE_LX("error: couldn't open %s\n", confFile);
    NiceExit(-1, "error: couldn't open conf file");
  }
  
  while (1) {
    /* get line */
    if (fgets(line, sizeof(line), list) == NULL) 
      break;

    /* end at any comment */
    if ((temp = strchr(line, '#')) != NULL)
      *temp = 0;

    if ((temp = GetField(line, 0)) == NULL)
      continue;
    
    /* version number */
    if (strncasecmp(temp, OPT_VERSION, sizeof(OPT_VERSION)-1) == 0) {
      ver = strtoll(GetField(temp, 1), NULL, 10);
      fclose(list); 
      return (ver > g_curConfigVersion);
    }
    
    /* change of the context */
    if (*temp == '*') {
      temp = GetField(temp, 1);
      if (strncasecmp(temp, OPT_NODLIST, sizeof(OPT_NODLIST)-1) == 0) 
	/* options part is over */
	break;
    }
  }
    
  fclose(list);
  return FALSE;
}
/*-----------------------------------------------------------------*/
void
ReadConfigFile(int readOptionsOnly)
{
  FILE *list;
  char confFile[MAX_PATH];
  PeerInfo* pi;
  char *addr;
  struct in_addr n_addr;
  int isOption = TRUE; /* start with options */
  int i;

  if (g_numPeers == 0) {
    PREPEND_BASEPATH(confFile, CONFIGFILE);
  } else {
    PREPEND_BASEPATH(confFile, NEWCONFIGFILE);
    if (access(confFile, R_OK) != 0) {
      /* if this file is not readable 
	 or do not exist, forget about it */
      TRACE_LX("attempt to reload a file that's not readable\n");
      return;
    }
  }

  if ((list = fopen(confFile, "r")) == NULL) {
    TRACE_LX("error: couldn't open %s\n", confFile);
    NiceExit(-1, "error: couldn't open conf file");
  }

  /* if it is reloading, free the name field */
  if (!readOptionsOnly && g_numPeers > 0) {
    for (i = 0; i < g_numPeers; i++) {
      pi = &g_allPeers[i];
      FREE_MEM(pi->pi_name);
    }
    g_numPeers = 0;
  }

  while (1) {
    char line[1024];
    char *temp;

    /* get one line */
    if (fgets(line, sizeof(line), list) == NULL) {
      fclose(list);
      break;
    }
    
    /* end at any comment */
    if ((temp = strchr(line, '#')) != NULL)
      *temp = 0;

    /* are we parsing options ? */
    if (isOption) {
      
      if ((temp = GetField(line, 0)) == NULL)
	continue;

      /* change of the context */
      if (*temp == '*') {
	temp = GetField(temp, 1);
	if (strncasecmp(temp, OPT_NODLIST, sizeof(OPT_NODLIST)-1) == 0) {
	  isOption = FALSE;
	  if (readOptionsOnly) {
	    fclose(list);
	    return;
	  }
	}
	continue;
      }

      /* initial timeout for UDP queries */     
      if (strncasecmp(temp, OPT_TIMEOUT, sizeof(OPT_TIMEOUT)-1) == 0) {
	g_udpTimeout = atoi(GetField(temp, 1)) * 1000;
	continue;
      }

      /* waiting period(secs) between heartbeats */     
      if (strncasecmp(temp, OPT_HB_TIMEOUT, sizeof(OPT_HB_TIMEOUT)-1) == 0) {
	g_HeartbeatTimeout = atoi(GetField(temp, 1));
	if (g_HeartbeatTimeout <= 0 || g_HeartbeatTimeout >= 300) {
	  TRACE_LX("invalid heartbeat was provided=%d, we reset it to one sec",
		   g_HeartbeatTimeout);
	  g_HeartbeatTimeout = 1;
	}
	continue;
      }

      /* number of slaves to be created */     
      if (strncasecmp(temp, OPT_NUM_SLAVES, sizeof(OPT_NUM_SLAVES)-1) == 0) {
	g_NumSlaves = atoi(GetField(temp, 1));
	if (g_NumSlaves < MIN_NUM_SLAVES) {
	  TRACE_LX("too small # of slaves=%d, reset it to min # of slaves=%d",
		  g_NumSlaves, MIN_NUM_SLAVES);
	  g_NumSlaves = MIN_NUM_SLAVES;
	}
	else if (g_NumSlaves > MAX_NUM_SLAVES) {
	  TRACE_LX("too big # of slaves=%d, reset it to max # of slaves=%d",
		  g_NumSlaves, MAX_NUM_SLAVES);
	  g_NumSlaves = MAX_NUM_SLAVES;
	}
	continue;
      }

      /* version number */
      if (strncasecmp(temp, OPT_VERSION, sizeof(OPT_VERSION)-1) == 0) {
	g_curConfigVersion = strtoll(GetField(temp, 1), NULL, 10);
	TRACE_LX("current config file ver=%lld\n", g_curConfigVersion);
	continue;
      }

      /* configuration file auto update URL */
      if (strncasecmp(temp, OPT_AUTOUPDATE, sizeof(OPT_AUTOUPDATE)-1) == 0) {
	FREE_MEM(g_autoUpdateHost);
	FREE_MEM(g_autoUpdateFile);
	g_autoUpdateHost = GetWord(temp, 1);
	g_autoUpdateFile = GetWord(temp, 2);
	TRACE_LX("AutoUpdate URL=%s%s\n", g_autoUpdateHost, g_autoUpdateFile); 
      }

      /* for safety */
      continue;
    }

#define DEFAULT_HOST "codeen.cs.princeton.edu"
#define DEFAULT_FILE "/codns/codns.conf"
    if (g_autoUpdateHost == NULL && g_autoUpdateFile == NULL) {
      g_autoUpdateHost = strdup(DEFAULT_HOST);
      g_autoUpdateFile = strdup(DEFAULT_FILE);
    }
    
    /* find the node name, copy it */
    if ((temp = GetWord(line, 0)) == NULL) 
      continue;
    
    /* find its IP address */
    memset(&n_addr, 0, sizeof(n_addr));
    if ((addr = GetField(line, 1)) != NULL)
      inet_aton(addr, &n_addr);
    
    /* initialize peer information */
    pi = &g_allPeers[g_numPeers];
    memset(pi, 0, sizeof(*pi));
    pi->pi_name = temp;
    pi->pi_address = n_addr;
    pi->pi_addrHash = WidespreadHash(HashString(temp, 0, FALSE));
    pi->pi_avgPingTime = INT_MAX;
    pi->pi_avgLocalLookupMS = INT_MAX;
    pi->pi_hbGoodAckHistory = hbAckHistoryMask;
    pi->pi_idx = g_numPeers;
    
    /* is this myself? */
    if (strcasecmp(pi->pi_name, g_localHostname) == 0)
      g_localHostIdx = g_numPeers;
    
    g_numPeers++;
    if (g_numPeers >= MAX_PEERS) /* it's enough */
      break;
  }

  /* init my neighbors */
  InitNeighborsList();
  
  /* maximum number of neighbors near myself */
  g_maxNeighbors = __min(g_numPeers, 30);
  TRACE_L("g_maxNeighbors = %d\n", g_maxNeighbors);
}
/*-----------------------------------------------------------------*/
static void
PeerLookupMain(int fd, void *token)
{
  /* go through the list of peers, use system to look them up
     the purpose of this process is
     1. initialize the address of each peer
     2. catch up with the IP address change of the peers */
  int i, sock, retry = 0;
  int initializing = TRUE;
  struct sockaddr_in sa;

  if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
    TRACE_LX( "failed creating socket in peer lookup\n");
    NiceExit(-1, "failed creating socket in peer lookup");
  }

  /* connect to local port */
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  sa.sin_port = htons(CODNS_PORT);

  while (1) {
    /* wait until the main process is ready */
    sleep(1); 

    if (connect(sock, (struct sockaddr *) &sa, sizeof(sa)) >= 0)
      break;

    if (retry++ >= 3) {
      TRACE_LX1( "too many failure in connecting\n");
      NiceExit(-1, "too many failure in connecting");
    }
    TRACE_LX("failed connecting in peer lookup, %d th\n", retry);
  }

  while (1) {
    int failure = 0;

    for (i = 0; i < g_numPeers; i++) {
      LocalQuery query;
      LocalQueryResult result;
      int len, res = 0;
      PeerInfo* pi;

      /* initialize to 0 */
      memset(&query, 0, sizeof(query));

      /* resolve peer names using CoDNS */
      pi = &g_allPeers[i];
      len = strlen(pi->pi_name) + 1;
      query.lq_info.lqi_size = len;         /* length of name */
      query.lq_info.lqi_id = INT_QUERY_ID;  /* we don't care */
      query.lq_info.lqi_cache = TRUE;
      strcpy(query.lq_name, pi->pi_name);
      len += sizeof(query.lq_info) + sizeof(query.lq_zero);

      /* send a query to CoDNS & receive the result of it */
      hton_LocalQueryInfo(&query.lq_info);
      if (write(sock, &query, len) != len) {
	TRACE_LX1( "failed writing out in peer lookup\n");
	NiceExit(-1, "failed writing out in peer lookup");
      }
      if ((res = read(sock, &result, sizeof(result))) != sizeof(result)) {
	TRACE_LX( "failed reading back in peer lookup : res = %d \n", res);
	NiceExit(-1, "failed reading back in peer lookup");
      }
      ntoh_LocalQueryResult(&result);

      /* skip the failure cases */
      if (result.lq_address[0].s_addr == 0) {
	TRACE_L("%s : lookup response is not valid\n", pi->pi_name);
	failure++;
      } 
      else if (pi->pi_address.s_addr != result.lq_address[0].s_addr) {
	/* convert it to a structure from the string, 
	   write it to the main process */
	PeerLookupResult resultAddr;
	
	/* update mine first */
	pi->pi_address = result.lq_address[0];

	/* send the result to the main process */
	resultAddr.plr_configVer = g_curConfigVersion;
	resultAddr.plr_idx = i;
	resultAddr.plr_address = result.lq_address[0];
	if (write(fd, &resultAddr, sizeof(resultAddr)) 
	    != sizeof(resultAddr)) {
	  TRACE_L1( "failed writing result in peer lookup\n");
	  NiceExit(-1, "failed writing result in peer lookup");
	}
      }

      /* slow it down a little */
      if (!initializing)
	sleep(1);	    

    }
    /* next round */
    initializing = FALSE;

    TRACE_L("%d failure out of %d\n", failure, g_numPeers);
    if (10 * failure < g_numPeers) {
      /* if failure rate is < 10 %, wait for ten hours + random alpha */
      int r = random() % 1801;

      TRACE_L("about to sleep %d secs\n", r+36000);
      sleep(r + 36000);                  
    }
    if (IsConfigFileUpdated(NULL)) {
      TRACE_L("config file updated\n");
      ReadConfigFile(FALSE);
    }
  }
  TRACE_L1("PeerLookupMain exited\n");
  NiceExit(0, "PeerLookupMain exited");
}
/*-----------------------------------------------------------------*/
static int
CreateUDPSocket(int port)
{
  /* creates an internet-domain udp socket on a well-known port, and
     returns its fd */
  int fd = CreatePublicUDPSocket(port);

  if (fd < 0) {
    TRACE_LX1( "failed making UDP socket\n");
    NiceExit(-1, "failed making UDP socket");
  }
  /* make it non blocking */
  if (fcntl(fd, F_SETFL, O_NDELAY) == -1) {
    TRACE_LX1("failed to make the socket non-blocking\n");
    close(fd);
    return(-1);
  }
  return(fd);
}
/*-----------------------------------------------------------------*/
static int
CreateTCPSocket(int port)
{
  /* create a local socket, have it start listen, return its fd */
  int fd;

  if ((fd = CreatePrivateAcceptSocketEx(port, TRUE, TRUE)) > 0)
    return(fd);

  TRACE_LX1( "failed creating accept port\n");
  NiceExit(-1, "failed creating accept port");
}
/*-----------------------------------------------------------------*/
static int
TimeCompare(struct timeval *a, struct timeval *b)
{
  /* returns -1, 0, or 1 if a is <, =, or > b */
  if (a->tv_sec < b->tv_sec ||
      (a->tv_sec == b->tv_sec && a->tv_usec < b->tv_usec))
    return(-1);
  if (a->tv_sec == b->tv_sec && a->tv_usec == b->tv_usec)
    return(0);
  return(1);
}
/*-----------------------------------------------------------------*/
static int
Before(QuerySource *entry, QuerySource *test)
{
  /* returns TRUE if we should insert entry before test */
  /* this entry goes first for one of three reasons:
     a) it's the only entry
     b) it's the only tcp entry
     c) it's the tcp entry with the lowest time
     */
  if (test == NULL ||
      test->qs_isUDP ||
      TimeCompare(&entry->qs_nextUDPTime, &test->qs_nextUDPTime) < 1)
    return(TRUE);
  return(FALSE);
}
/*-----------------------------------------------------------------*/
static void
AddToAll(QuerySource *qs, int removeFirst)
{
  /* adds it to the list of all queries, in sorted order. if the
     removeFirst parameter is specified, the entry is first removed
     from the list and then added to the appropriate place*/
  QuerySource *walk;

  if (removeFirst)
    TAILQ_REMOVE(&allQueries, qs, qs_all);

  if (qs->qs_isUDP && qs->qs_nextUDPTime.tv_sec) {
    TRACE_LX1( "had udp query with udp time\n");
    NiceExit(-1, "had udp query with udp time");
  }

  if (qs->qs_isUDP) {
    /* add it to the end - we don't do udp queries on received queries */
    TAILQ_INSERT_TAIL(&allQueries, qs, qs_all);
    return;
  }

  walk = TAILQ_FIRST(&allQueries);
  if (Before(qs, walk)) {
    TAILQ_INSERT_HEAD(&allQueries, qs, qs_all);
    return;
  }

  /* walk the list until we're in sorted order, or until
     we hit the udp entries */
  TAILQ_FOREACH(walk, &allQueries, qs_all) {
    if (Before(qs, walk)) {
     TAILQ_INSERT_BEFORE(walk, qs, qs_all);
     return;
    }
    if (TAILQ_NEXT(walk, qs_all) == NULL) {
      TAILQ_INSERT_TAIL(&allQueries, qs, qs_all);
      return;
    }
  }

  TRACE_LX1( "walked past end in AddToAll\n");
  NiceExit(-1, "walked past end in AddToAll");
}
/*-----------------------------------------------------------------*/
static void
AddToQueue(QuerySource *qs, struct WaitList *list)
{
  /* add the request to the specified queue, and to the queue
     containing all requests */
  /* TRACE_L( "entering %s\n", __FUNCTION__); */

  AddToAll(qs, FALSE);
  LIST_INSERT_HEAD(list, qs, qs_waiting);
  gettimeofdayex(&qs->qs_startTime, NULL);
  qs->qs_onWaiting = TRUE;
}
/*-----------------------------------------------------------------*/
static void
SetUDPDelay(QuerySource *qs, int microsecs)
{
#define MIN_DELAY (.05*1e6)
#define MAX_DELAY (2*1e6)
  if (microsecs == 0) 
    microsecs = ((int)GetMostPeerRespTime()*1000);
  if (microsecs < MIN_DELAY)
    microsecs = MIN_DELAY;
  else if (microsecs > MAX_DELAY)
    microsecs = MAX_DELAY;

  qs->qs_udpDelayUsec = microsecs;
  qs->qs_nextUDPTime = g_currentTime;
  qs->qs_nextUDPTime.tv_usec += microsecs;
  if (qs->qs_nextUDPTime.tv_usec >= 1000000) {
    qs->qs_nextUDPTime.tv_sec += qs->qs_nextUDPTime.tv_usec / 1000000;
    qs->qs_nextUDPTime.tv_usec = qs->qs_nextUDPTime.tv_usec % 1000000;
  }
}
/*-----------------------------------------------------------------*/
static void
SendRemoteQueries(QuerySource *qs, int fd)
{
  struct sockaddr_in addr;
  int serv, delay;
  struct timeval currTime;

  /* double the delay time, and adjust our position on list.  do this
     before seeing if we actually sent, since we still want to delay
     if we can't send */
  SetUDPDelay(qs, 2 * qs->qs_udpDelayUsec);
  AddToAll(qs, TRUE);

  if (qs->qs_fd < 0) {
    /* we're on the queue only because the g_slave is still working
       on it, while the response has already been sent via another
       mechanism. So, don't bother sending remote queries */
    return;
  }

  /* the exact delay from the request arrival time */
  gettimeofdayex(&currTime, NULL);
  delay = (int)DiffTimesToMSec(currTime, qs->qs_startTime);

  /* pick a peer node for this remote query */
  serv = PR_HRWPickServer(qs->qs_query.nsq_name);
  if (serv < 0) {
    TRACE_L("%09d.%03d %s %d :couldn't find a good peer DNS\n", 
	    (int)qs->qs_startTime.tv_sec, 
	    (int)(qs->qs_startTime.tv_usec /1000),
	    qs->qs_query.nsq_name,
	    delay);
    return;
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_addr = g_allPeers[serv].pi_address;
  addr.sin_port = htons(CODNS_PORT);

  /* record some bookkeeping information */
  qs->qs_numUDPs++;
  qs->qs_query.nsq_startTime = qs->qs_startTime; /* for future analysis */

  /* the exact delay from the request arrival time */
  gettimeofdayex(&currTime, NULL);
  qs->qs_query.nsq_delay = delay;

  hton_NSQuery(&qs->qs_query);
  if (sendto(fd, &qs->qs_query, sizeof(NSQuery), 0,
	     (struct sockaddr *) &addr,
	     sizeof(struct sockaddr_in)) != sizeof(NSQuery)) {

    TRACE_L("%09d.%03d %s %d: weird peer sendto size\n",
	    (int)qs->qs_startTime.tv_sec, 
	    (int)(qs->qs_startTime.tv_usec /1000),
	    qs->qs_query.nsq_name,
	    delay);
    ntoh_NSQuery(&qs->qs_query);
  } else {
    /* leave a log for remote query */
    ntoh_NSQuery(&qs->qs_query);
    RecordRemoteQuery(qs, serv);
  }

}

/*-----------------------------------------------------------------*/
static void
SendRequestToSlave(QuerySource *qs)
{
  /* pull request off waiting list. mark request as waiting. pull a
     free g_slave off the list. send the request to the g_slave */
  SlaveInfo *si = LIST_FIRST(&freeSlaves);
  int len = strlen(qs->qs_query.nsq_name)+1;

  if (si == NULL) {
    TRACE_LX1( "SendRequestToSlave had no slaves\n");
    NiceExit(-1, "SendRequestToSlave had no slaves");
  }

  LIST_REMOVE(si, si_free);
  LIST_REMOVE(qs, qs_waiting);
  qs->qs_onWaiting = FALSE;
  qs->qs_atSlave = TRUE;
  qs->qs_hitSlave = 1;
  si->si_qs = qs;
  si->si_inUse = 1;
  g_slaveInUse++;
  if (write(si->si_sock, qs->qs_query.nsq_name, len) != len) {
    TRACE_LX1( "couldn't write full name to g_slave\n");
    NiceExit(-1, "couldn't write full name to g_slave");
  }
}

/*-----------------------------------------------------------------*/
static void
WriteConfigFile(void) 
{ 
  /* periodically back up config files */
  FILE *fp;
  PeerInfo* pi;
  int pi_idx[MAX_PEERS];
  int i, num, pid;
  char confFile[MAX_PATH];
  char tempFile[MAX_PATH];
  struct stat buf;

  PREPEND_BASEPATH(confFile, CONFIGFILE);
  PREPEND_BASEPATH(tempFile, CONFIGFILE ".temp");

  /* we fork a new process for saving ip list
     to prevent blocking while doing file i/o */
  if ((pid = fork()) < 0) {
    TRACE_LX1("failed on fork\n");
    NiceExit(-1, "failed on fork");
  }

  if (pid > 0) {
    /* we're in the parent */
    return;
  }

  /* we're in a child */
  if ((fp = fopen(tempFile, "w")) == NULL) {
    TRACE_LX("error: couldn't open %s\n", tempFile);
    goto done;
  }

  /* copy the option part */
  fprintf(fp, "# options followed by nodelist\n\n");

  /* version and update host */
  fprintf(fp, "# please don't change this part\n");
  fprintf(fp, "%s %lld\n", OPT_VERSION, g_curConfigVersion);
  fprintf(fp, "%s %s %s\n\n", 
	  OPT_AUTOUPDATE, g_autoUpdateHost, g_autoUpdateFile);

  /* initial timeout */
  fprintf(fp, "# initial timeout for udp queries\n");
  fprintf(fp, "%s %d # %d ms\n\n", 
	  OPT_TIMEOUT, g_udpTimeout/1000, g_udpTimeout/1000);

  if (g_HeartbeatTimeout > 1) {
    /* heartbeat timeout */
    fprintf(fp, "# waiting periods(secs) between heartbeats\n");
    fprintf(fp, "%s %d # %d secs\n\n", 
	    OPT_HB_TIMEOUT, g_HeartbeatTimeout, g_HeartbeatTimeout);
  }

  if (g_NumSlaves != MAX_NUM_SLAVES) {
    /* number of slaves to be created */
    fprintf(fp, "# number of slaves to be created\n");
    fprintf(fp, "%s %d\n\n", OPT_NUM_SLAVES, g_NumSlaves);
  }

  /* write node list & its IPs according to their order */
  fprintf(fp, "# list of node ip pairs\n");
  fprintf(fp, "# this part must start with \"* nodelist\"\n");
  fprintf(fp, "* nodelist\n");
  GetNeighborsSnapshot(pi_idx, &num, TRUE);
  for (i = 0; i < num; i++) {
    pi = &g_allPeers[pi_idx[i]];
    fprintf(fp, "%s %s\n", pi->pi_name, inet_ntoa(pi->pi_address));
  }
  fclose(fp);

  /* only when there's some contents, we rename it */
  if (stat(tempFile, &buf) == 0 && buf.st_size > 0) {
    /* rename it to conf file */
    if (rename(tempFile, confFile) < 0) {
      TRACE_LX("rename failed reason =%s\n", strerror(errno));
    }
  } else {
    TRACE_LX("tempFile size is 0\n");
  }
  unlink(tempFile);

  /* the child process should terminate here */
done:
  exit(0);
}
/*-----------------------------------------------------------------*/
static int
IsGlobalTimeout(void)
{
  return (DiffTimesToMSec(g_nextGlobalTimeout, g_currentTime) < 0);
}
/*-----------------------------------------------------------------*/
void
OnGlobalTimerExpire(int hbSock)
{
  static struct timeval time_span = {1, 0};        /* 1 sec */

  /* set time_span if a new value is obtained in the config file */
  if (time_span.tv_sec != g_HeartbeatTimeout)
    time_span.tv_sec = g_HeartbeatTimeout;

  /* send heart beats */
  SendHeartbeatQueries(hbSock);

  /* set next timeout */
  g_nextGlobalTimeout = AddTimes(g_currentTime, time_span);

  /* periodically save ip list for nodes */
  if ((timeex(NULL) % 3600) == 0) {
    if (IsConfigFileUpdated(NULL)) {
      TRACE_L("config file updated\n");
      ReadConfigFile(FALSE);
    }
    WriteConfigFile();
  }
}
/*-----------------------------------------------------------------*/
static void
Init_LogFile(void)
{
  char debugFile[MAX_PATH], logDir[MAX_PATH], exitLogFile[MAX_PATH];
  struct stat buf;

  /* exit log file */
  PREPEND_BASEPATH(exitLogFile, "exitlog");
  if (NiceExitOpenLog(exitLogFile) != SUCCESS) {
    fprintf(stderr, "failed opening exitlog");
    exit(-1);
  }

  PREPEND_BASEPATH(debugFile, "dbgcodns");

  /* see if we have "log" directory */
  PREPEND_BASEPATH(logDir, "log");
  if (stat(logDir, &buf) == 0 && buf.st_mode & S_IFDIR) {
    PREPEND_BASEPATH(debugFile, "log/dbgcodns");
  }

  hdebugLog = CreateLogFHandle(debugFile, TRUE);
  if (hdebugLog == NULL)
    hdebugLog = CreateLogFHandle("dbgcodns", TRUE);
  if (hdebugLog == NULL || (OpenLogF(hdebugLog) < 0)) {
    fprintf(stderr, "can't do it\n");
    NiceExit(-1, "failed opening log");
  }
}
/*-----------------------------------------------------------------*/
static void
CheckPeerLookup(const SelCtx* ctx) 
{ 
  PeerLookupResult result;
  PeerInfo* pi;

  /* see if we've got some peer address translations */
  if (!FD_ISSET(ctx->sel_fd, ctx->sel_tempSet)) 
    return;

  if (read(ctx->sel_fd, &result, sizeof(result)) != sizeof(result)) {
    TRACE_LX1( "weird read in translation\n");
    NiceExit(-1, "weird read in translation");
  }

  /* skip when two processes see different version of config file */
  if (g_curConfigVersion != result.plr_configVer)
    return;

  /* out of valid index */
  if (result.plr_idx >= g_numPeers) {
    TRACE_L("result index out of the limit %d\n", result.plr_idx);
    return;
  }

  /* update my peer address */
  pi = &g_allPeers[result.plr_idx];
  if (pi->pi_address.s_addr != result.plr_address.s_addr)
    pi->pi_address = result.plr_address;
  else
    TRACE_L("unnecessary ip address update for %s\n", pi->pi_name);
}

/*-----------------------------------------------------------------*/
static void
CheckSlave(const SelCtx* ctx)
{
/* if the lookup is for internal use/codeen's check
   then it doesn't reflect the win history */
#define IS_INTERNAL_USE(x) (((x)->qs_queryinfo.lqi_id == INT_QUERY_ID) || \
                            ((x)->qs_queryinfo.lqi_id == HBTWGET_CODNS_ID))

  int i;

  /* see if any g_slaves are done */
  for (i = 0; i < g_NumSlaves; i++) {
    SlaveInfo *si = &g_slave[i];
    int res, size;

    /* skip if no answer from this slave */
    if (!FD_ISSET(si->si_sock, ctx->sel_tempSet)) 
      continue;

    /* slave not in use responded */
    if (!si->si_inUse) {
      TRACE_LX1( "slave not in use responded\n");
      NiceExit(-1, "slave not in use responded");
    }

    /* get a response */
    size = sizeof(NSQueryResult);
    if ((res = read(si->si_sock, &si->si_qs->qs_query.nsq_result, size))
	!= size) {
      TRACE_LX("slave gave weird result - %d, %d\n", res, size);
      NiceExit(-1, "slave gave weird result");
    }

    /* if this is for my local name lookup, then update the statistics */
    if (!si->si_qs->qs_isUDP)
      UpdateDNSStatistics(si->si_qs, TRUE);

    /* record win's history 
       we ignore internal name lookups and zero answer cases */
    if (!IS_INTERNAL_USE(si->si_qs) && 
	si->si_qs->qs_query.nsq_result.nsq_address[0].s_addr != 0)
      g_udpWinsHistory <<= 1;

    if (SendAllResponse(si->si_qs, ctx->sel_masterSet)) {
      if (!IS_INTERNAL_USE(si->si_qs) &&
	  si->si_qs->qs_query.nsq_result.nsq_address[0].s_addr != 0)
	g_udpWinsHistory |= 1;

      si->si_qs->qs_slaveLost = 1;
    }

    FreeSlave(i);
  }
}
/*-----------------------------------------------------------------*/
static int
HasRemoteQuery(const QuerySource* family, struct sockaddr_in origin)
{
  /* see if my family has already got a UDP request 
     from the same origin */
  QuerySource* walk;

  if (family->qs_isUDP && 
      family->qs_origin.sin_addr.s_addr == origin.sin_addr.s_addr)
    return TRUE;

  LIST_FOREACH(walk, &family->qs_familyList, qs_family) {
    if (walk->qs_isUDP && 
	walk->qs_origin.sin_addr.s_addr == origin.sin_addr.s_addr) 
      return TRUE;
  }
  return FALSE;
}
/*-----------------------------------------------------------------*/
static void
CheckUDPSocket(const SelCtx* ctx)
{
  QuerySource* query;

  /* see if we've got requests/responses from other nodes */
  if (!FD_ISSET(ctx->sel_fd, ctx->sel_tempSet))
    return;

  /* read a query */
  if ((query = ReadQuery(ctx->sel_fd, TRUE, NULL)) == NULL)
    return;

  if ((query->qs_query.nsq_result.nsq_address)[0].s_addr == 0) {
    /* we have a UDP query request from a peer */
    QuerySource *family;

    /* see if my family exists */
    if ((family = QueryMatch(query, FALSE, TRUE)) != NULL) {

      /* if my family has alreay got an answer, send it right away 
	 as an response. it's the case where a remote query won but 
	 CoDNS still waits for the local DNS response */
      if (family->qs_fd < 0) {
	query->qs_query.nsq_result = family->qs_query.nsq_result;
	SendResponse(query, NULL, NULL);
	FREE_MEM(query);
      } 
      else if (HasRemoteQuery(family, query->qs_origin)) {
	/* if we've aleady got the same request from the same origin,
	   we just forget about it */
	TRACE_L("already got the request(%s) from %s\n",
	      query->qs_query.nsq_name, inet_ntoa(query->qs_origin.sin_addr));
	FREE_MEM(query);
      } 
      else {
	/* add to my family list */
	LIST_INSERT_HEAD(&family->qs_familyList, query, qs_family);
      }
    } 
    else {
      /* now, let's queue it */
      if (g_udpWaitSize < MAX_UDP_WAITSIZE 
	  && g_slaveInUse < g_NumSlaves/2) {
	AddToQueue(query, &udpWait);
	g_udpWaitSize++;
      } else {
	/* no point storing too many udp queries */
	FREE_MEM(query);	
      }
    }
  }
  else {
    /* it's a response - check if we still have original query. 
       if it's not at the g_slave, free it */
    QuerySource *orig;

    orig = QueryMatch(query, TRUE, FALSE);

    /* update DNS statistics */
    query->qs_startTime = query->qs_query.nsq_startTime;
    query->qs_fd = (orig == NULL) ? -1 : orig->qs_fd;
    if (orig != NULL) {
      query->qs_queryinfo = orig->qs_queryinfo;
    }
    UpdateDNSStatistics(query, FALSE);

    if (orig != NULL) {    
      if (orig->qs_fd >= 0) {
	orig->qs_query.nsq_result = query->qs_query.nsq_result;
	SendAllResponse(orig, ctx->sel_masterSet);
	if (!orig->qs_atSlave)
	  FreeQuery(orig);
      }
    }
    FREE_MEM(query);		/* not on any lists, so easy to free */
  }
}
/*-----------------------------------------------------------------*/
static void
CheckLocalQueries(const SelCtx* ctx)
{
  /* we've got new local queries */
  PersistConnFD* walk;
  QuerySource* query;
  int delay;
  LIST_HEAD(, PersistConnFD) deadList = LIST_HEAD_INITIALIZER(deadList);

  LIST_FOREACH(walk, &allFDs, pc_all) {
    int need_close = FALSE;

    /* if this fd is not set, skip it */
    if (!FD_ISSET(walk->pc_fd, ctx->sel_tempSet)) 
      continue;

    /* read a query */
    if ((query = ReadQuery(walk->pc_fd, FALSE, &need_close)) != NULL) {
      /* we've got a new query */      
       QuerySource* family;

      /* see if my family already exists */
      if ((family = QueryMatch(query, FALSE, FALSE)) != NULL) {
	/* if my family has already got an answer, send it right away 
	   as an response. it's the case where a remote query won but 
	   CoDNS still waits for the local DNS response */

	if (family->qs_fd < 0) {
	  query->qs_query.nsq_result = family->qs_query.nsq_result;
	  SendResponse(query, NULL, &need_close);
	  FREE_MEM(query);
	} else {
	  /* attach me to my family list */
	  LIST_INSERT_HEAD(&family->qs_familyList, query, qs_family);
	}

      } else {
	/* I'm a new query with no family */

	delay = GetUDPStartDelay(g_udpWinsHistory, g_udpTimeout, MAX_DELAY);
	SetUDPDelay(query, delay);
	AddToQueue(query, &tcpWait);
	if (delay == 0) { /* this means our local DNS is pretty bad */
	  query->qs_udpDelayUsec = 0; /* don't count this send */
	  SendRemoteQueries(query, ctx->sel_fd);
	}
      }
    }

    /* see if the connetion needs to be closed &
       add it to the dead list for future cleanup */
    if (need_close)   
      LIST_INSERT_HEAD(&deadList, walk, pc_dead);
  }
  
  /* clean up the dead entries */
  while ((walk = LIST_FIRST(&deadList)) != NULL) {
    FD_CLR(walk->pc_fd, ctx->sel_masterSet);
    close(walk->pc_fd);

    LIST_REMOVE(walk, pc_all);
    LIST_REMOVE(walk, pc_dead);
    FREE_MEM(walk);
  }
}
/*-----------------------------------------------------------------*/
static void
CheckLocalConnection(const SelCtx* ctx)
{
  int new_fd;

  /* see if we've got any new local connect attempts */
  if (!FD_ISSET(ctx->sel_fd, ctx->sel_tempSet))
    return;

  if ((new_fd = accept(ctx->sel_fd, NULL, NULL)) < 0) {
    TRACE_LX1( "accept had error\n");
    NiceExit(-1, "accept had error");
  }
  
  /* make it non-blocking */
  if (fcntl(new_fd, F_SETFL, O_NDELAY) < 0) {
    TRACE_LX1("fcntl failed\n");
    close(new_fd);
  } else {
    if (AddPersistConn(new_fd) < 0) {
      TRACE_LX1("adding to persistent connetion list failed\n");
      NiceExit(-1, "adding to persistent connetion list failed");
    }
    FD_SET(new_fd, ctx->sel_masterSet);
  }
}
/*-----------------------------------------------------------------*/
static void
SigChildHandler(int sig)
{
  int pid;

 do {
    pid = waitpid(-1, NULL, WNOHANG);
  } while (pid > 0);
}
/*-----------------------------------------------------------------*/
static void
DefaultSigHandler(int sig)
{
  char buf[128];

  sprintf(buf, "caught sig %d - exiting", sig);
  NiceExit(-1, buf);
}
/*-----------------------------------------------------------------*/
static void
FillBasePath(void)
{
  const char* envPath = NULL;

  if ((envPath = getenv("CODNS_PATH")) != NULL) {
    int len;

    snprintf(g_basePath, sizeof(g_basePath), "%s", envPath);
    len = strlen(g_basePath);
    if (len > 0 && g_basePath[len-1] == '/')
      g_basePath[len-1] = 0;

  } else {
    char codnsPath[MAX_PATH];

    if (getcwd(g_basePath, sizeof(g_basePath)) == NULL) {
      fprintf(stderr, "getcwd failed\n");
      exit(-1);
    }

    PREPEND_BASEPATH(codnsPath, "codns");
    if (access(codnsPath, F_OK) < 0) {
      fprintf(stderr, "please execute codns in the same directory as"
	      "its binary");
      exit(-1);
    }
  }
}
/*-----------------------------------------------------------------*/
int
main(int argc, char *argv[])
{
#define EARLIER(x, y) (DiffTimesToMSec((x), (y)) < 0 ? (x) : (y))
#define CHECK_TIME0(idx, max) {                                      \
      struct timeval now;                                            \
      int msDiff;                                                    \
      gettimeofdayex(&now, NULL);                                    \
      msDiff = (int)DiffTimesToMSec(now, from);                      \
      if (msDiff >= (max)) {                                         \
	TRACE_L("too slow %d : %d ms\n", (idx), msDiff);             \
      }                                                              \
      from = now;                                                    \
    }
#define CHECK_TIME(idx) CHECK_TIME0(idx, 1000)
#define CALL_CHECKFUNC(fd, func) \
      selctx.sel_fd = fd;        \
      func(&selctx);

  int i;
  int udpSocket;
  int localDomainSocket;
  int peerLookupFD;
  int hbSock;
  int statQuerySocket;
  fd_set masterSet, tempSet;
  SelCtx selctx;
  struct sigaction act;

  /* know myself first */
  if (gethostname(g_localHostname, sizeof(g_localHostname)) < 0) {
    fprintf(stderr, "couldn't get local hostname\n");
    exit(-1);
  }

  /* fill base path */
  FillBasePath();
  
  /* inialize log file setup */
  Init_LogFile();

  /* create log slaves */
  g_logFd = CreateSlave(LogSlaveMain, (void *)1, FALSE);
  if (fcntl(g_logFd, F_SETFL, O_NDELAY) < 0) {
     fprintf(stderr, "logFd: fcntl failed\n");
     NiceExit(-1, "logFd: fcntl failed");
  }

  TRACE_L("parent pid is %d\n", (int) getpid());

  /* establish a signal handler for SIGCHLD - just out of paranoia, do
     it after the fork. */
  sigaction(SIGCHLD, NULL, &act);
  if (act.sa_handler == SIG_IGN) {
    fprintf(stderr, "SIGCHILD is ignored\n");
    NiceExit(-1, "SIGCHILD is ignored");
  }
  act.sa_handler = SigChildHandler;
  sigemptyset (&act.sa_mask);
  act.sa_flags = SA_NOCLDSTOP | SA_RESTART;
  sigaction(SIGCHLD, &act, NULL);

  /* initialize peers */
  ReadConfigFile(FALSE);

  /* initialize random seed */
  gettimeofdayex(&g_currentTime, NULL);
  srandom(g_currentTime.tv_usec);
  FD_ZERO(&masterSet);

  /* create all g_slaves */
  for (i = 0; i < g_NumSlaves; i++) {
    if ((g_slave[i].si_sock = CreateSlave(DNSSlaveMain, NULL, TRUE)) < 0) {
      TRACE_LX1( "failed to create all g_slaves\n");
      NiceExit(-1, "failed to create all g_slaves");
    }
    FD_SET(g_slave[i].si_sock, &masterSet);
    LIST_INSERT_HEAD(&freeSlaves, &g_slave[i], si_free);
  }

  /* do signals after slaves, so that a killall doesn't generate
     many lines in exitlog */
  signal(SIGHUP, DefaultSigHandler);
  signal(SIGINT, DefaultSigHandler);
  signal(SIGPIPE, DefaultSigHandler);
  signal(SIGALRM, DefaultSigHandler);
  signal(SIGTERM, DefaultSigHandler);
  signal(SIGUSR1, DefaultSigHandler);
  signal(SIGUSR2, DefaultSigHandler);
#ifndef OS_DARWIN
  signal(SIGPOLL, DefaultSigHandler);
#else
  signal(SIGIO, DefaultSigHandler);
#endif
  signal(SIGPROF, DefaultSigHandler);
  signal(SIGVTALRM, DefaultSigHandler);

  /* open up UDP socket for outside queries (unreliable) */
  udpSocket = CreateUDPSocket(CODNS_PORT);
  FD_SET(udpSocket, &masterSet);

  /* open up a local socket for local queries (reliable) */
  localDomainSocket = CreateTCPSocket(CODNS_PORT);
  FD_SET(localDomainSocket, &masterSet);

  /* init the peer info after we've set up sockets, since the peer
     lookup will try to use them */
  peerLookupFD = CreateSlave(PeerLookupMain, NULL, FALSE);
  FD_SET(peerLookupFD, &masterSet);

  /* config file updater */
  CreateConfigFileAutoUpdater();

  /* set up heart beat sockets */
  hbSock = Heartbeat_Init();
  FD_SET(hbSock, &masterSet);

  /* open up UDP socket for statitics query */
  statQuerySocket = CreateTCPSocket(CODNS_STAT_PORT);
  FD_SET(statQuerySocket, &masterSet);

  /* fire first timeout to set up next timeout value */
  OnGlobalTimerExpire(hbSock);

  while (1) {
    struct timeval timeout;
    struct timeval *timeoutPtr = &timeout;
    QuerySource *query;
    struct timeval from;

    /* adjust the select timeout for select for the next global timeout */
    gettimeofdayex(&g_currentTime, NULL);
    if ((query = TAILQ_FIRST(&allQueries)) == NULL || query->qs_isUDP)
      timeout = DiffTimes(g_nextGlobalTimeout, g_currentTime);
    else
      timeout = DiffTimes(EARLIER(g_nextGlobalTimeout, query->qs_nextUDPTime),
			  g_currentTime);
    /* if negative, make it zero */
    if (IsTimeNegative(timeout)) 
      memset(&timeout, 0, sizeof(timeout));
    
    /* call select */
    from = g_currentTime;
    tempSet = masterSet;
    while (select(FD_SETSIZE, &tempSet, NULL, NULL, timeoutPtr) < 0) {
      if (errno == EINTR) {
	TRACE_L1("EINTR happened\n");
	continue;
      }
      TRACE_LX("error in select - %d\n", errno);
      perror("select");
      NiceExit(-1, "select error");
    }

    gettimeofdayex(&g_currentTime, NULL);
    if (timeout.tv_sec != 0 || timeout.tv_usec != 0) {
      CHECK_TIME0(0, (ToMSec(timeout) + 1000));
    } else {
      from = g_currentTime;
    }

    /* initialze select context */
    selctx.sel_tempSet = &tempSet;
    selctx.sel_masterSet = &masterSet;

    /* time out occurred */
    if (IsGlobalTimeout())
      OnGlobalTimerExpire(hbSock);

    CHECK_TIME(1);

    /* see if we've got heartbeat */
    if (FD_ISSET(hbSock, &tempSet))
      Heartbeat_Receive(hbSock, NULL);

    CHECK_TIME(2);

    CALL_CHECKFUNC(peerLookupFD, CheckPeerLookup);

    CHECK_TIME(3);

    CALL_CHECKFUNC(-1, CheckSlave);

    CHECK_TIME(4);

    CALL_CHECKFUNC(udpSocket, CheckUDPSocket);

    CHECK_TIME(5);

    CALL_CHECKFUNC(udpSocket, CheckLocalQueries);

    CHECK_TIME(6);

    CALL_CHECKFUNC(localDomainSocket, CheckLocalConnection);

    CHECK_TIME(7);

    CALL_CHECKFUNC(statQuerySocket, CheckStatQuery);

    CHECK_TIME(8);

    /* send local queries to slaves 
       before handling remote queries */
    while (LIST_FIRST(&freeSlaves) != NULL &&
	   LIST_FIRST(&tcpWait) != NULL)
      SendRequestToSlave(LIST_FIRST(&tcpWait));
    while (LIST_FIRST(&freeSlaves) != NULL &&
	   LIST_FIRST(&udpWait) != NULL) {
      SendRequestToSlave(LIST_FIRST(&udpWait));
      g_udpWaitSize--;
    }

    /* while any queries need udp sends,
       send the udps */
    while ((query = TAILQ_FIRST(&allQueries)) != NULL &&
	   (!query->qs_isUDP) &&
	   TimeCompare(&query->qs_nextUDPTime, &g_currentTime) < 1) {
      SendRemoteQueries(query, udpSocket);
    }

    CHECK_TIME(9);
  }
}
/*-----------------------------------------------------------------*/
