#ifndef _CODNS_INT_H_
#define _CODNS_INT_H_

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "codnslib.h"
#include "codns.h"
#include "queue.h"

/* CoDNS internal header 
   defines structures and variables for internal use only */

/*-------------------------------------------------------------------------*/
#define MAX_PEERS 600
typedef struct PeerInfo {
  char          *pi_name;	      /* name of peer node */
  struct in_addr pi_address;          /* address */
  unsigned int   pi_addrHash;         /* address hash value */
  int            pi_pingCount;        /* hbt counts*/
  int            pi_avgPingTime;      /* RTT from this node */
  int            pi_avgLocalLookupMS; /* average local DNS latency */
  time_t         pi_hbLastSentTime;   /* last time when a hbt was sent */
  int            pi_hbGoodAckHistory; /* hbt response history */
  int            pi_idx;              /* index in the array */
  struct in_addr pi_nameServers[2];   /* name server configurations */
  TAILQ_ENTRY(PeerInfo) node_all;     /* neighbor node list */
} PeerInfo;

/*-------------------------------------------------------------------------*/
/* some data structures for queries */
typedef struct NSQueryResult {
  struct in_addr nsq_address[MAX_ANSWERS]; /* result - IP address of name */
  int  nsq_ttl;                            /* TTL */
} NSQueryResult;

/*-------------------------------------------------------------------------*/

#define MAX_NSQUERY_NAME 256
typedef struct NSQuery {
  int nsq_version;		     /* version # */
  NSQueryResult nsq_result;	     /* result from slave */
  int nsq_randomToken;		     /* weak defense against attacks */
  char nsq_name[MAX_NSQUERY_NAME];   /* domain name */
  int nsq_delay;                     /* delay at the time of peer query */
  int nsq_id;                        /* original query id */
  struct timeval nsq_startTime;      /* the request arrival time */
} NSQuery;

/*-------------------------------------------------------------------------*/

typedef struct QuerySource {
  int qs_fd;			/* which fd should get response, -1 if done */
  int qs_isUDP;			/* was origin a UDP socket? */
  int qs_atSlave;	        /* is a slave currently handling? */
  int qs_onWaiting;
  int qs_udpDelayUsec;		/* backoff count in microseconds */
  int qs_numUDPs;		/* how many udps were sent */
  int qs_hitSlave;		/* did the slave service it */
  int qs_slaveLost;		/* did the slave answer first */
  LocalQueryInfo qs_queryinfo;  /* query info */
  NSQuery qs_query;           
  struct sockaddr_in qs_origin;	/* if remote, querying address */
  struct timeval qs_nextUDPTime;/* next time to send out a UDP query */
  struct timeval qs_startTime;  /* the starting time for name lookup */
  TAILQ_ENTRY(QuerySource) qs_all; /* all queries */
  LIST_ENTRY(QuerySource) qs_waiting; /* only queries waiting on slave */
  LIST_HEAD(FamilyList, QuerySource) qs_familyList; /* head of my family */
  LIST_ENTRY(QuerySource) qs_family; /* list of the queries of the same name */
} QuerySource;

/*-------------------------------------------------------------------------*/

typedef struct SelCtx {        /* select context */
  int     sel_fd;              /* file descriptor of interest */
  fd_set *sel_tempSet;         /* temp fd set */
  fd_set *sel_masterSet;       /* master fd set */
} SelCtx;

/*-------------------------------------------------------------------------*/
/* new gethostbyname() with ttl export */

extern int _dns_ttl_;
struct hostent *gethostbynameX(const char *name);

/*-------------------------------------------------------------------------*/
/* network <-> host byte order, defined in codnslib.c */

void ntoh_NSQueryResult(NSQueryResult* p);
void hton_NSQueryResult(NSQueryResult* p);
void ntoh_NSQuery(NSQuery* p);
void hton_NSQuery(NSQuery* p);
void ntoh_LocalQueryInfo(LocalQueryInfo *p);
void hton_LocalQueryInfo(LocalQueryInfo *p);
void ntoh_LocalQueryResult(LocalQueryResult* p);
void hton_LocalQueryResult(LocalQueryResult* p);

/*-------------------------------------------------------------------------*/
/* internal query ID for resolving peer nodes 
   CoDNS doesn't leave a log for the queries with this ID */

#define MAX_NEIGHBORS 10   /* max number of neighbor nodes */

#define INT_QUERY_ID -2
#define hbAckHistoryMask  (0x00000fff)
#define hbSendHistoryMask (0x00000001)

/* configuration file name */
#define CONFIGFILE    "codns.conf"
#define NEWCONFIGFILE CONFIGFILE ".new"

/* max path length */
#define MAX_PATH 1024

extern PeerInfo g_allPeers[];
extern int g_numPeers;
extern int g_localHostIdx;
extern int g_avgLocalLookupMS;
extern int g_maxNeighbors;
extern struct timeval g_currentTime;
extern char g_localHostname[];

#endif
