#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <ctype.h>
#include "codnshbt.h"
#include "codns_int.h"
#include "HRW_codns.h"
#include "debug.h"
#include "codnslog.h"
#include "ports.h"

#define VALID_PING_COUNT    50

/* numeric forms of the first two name servers in my configuration file */
static struct in_addr g_MyNameServers[2];

/*-----------------------------------------------------------------*/
static void
Heartbeat_Send(int sock, struct in_addr sin_addr, int portNum,
	       int status, int id, struct timeval* psendTime)
{
  /* send a udp heartbeat to a peer */
  CoDNSHeartbeatInfo hb;
  struct sockaddr_in addr;
  int res;

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_addr = sin_addr;
  addr.sin_port = htons(portNum);
  
  hb.hi_versionNum = CODNS_HBI_VERSION;
  hb.hi_statusVec = status;
  hb.hi_ID = id;
  hb.hi_avgLocalLookupMS = g_avgLocalLookupMS;
  hb.hi_nameServers[0] = g_MyNameServers[0];
  hb.hi_nameServers[1] = g_MyNameServers[1];
  if (psendTime != NULL)
    hb.hi_sendTime = *psendTime;
  else
    gettimeofdayex(&hb.hi_sendTime, NULL);

  hton_CoDNSHeartbeatInfo(&hb);
  res = sendto(sock, &hb, sizeof(hb), 0,
	       (struct sockaddr *) &addr, sizeof(addr));
  if (res < 0) {
    TRACE_L("Error in send to %s:%d for %s res: %d at socket %d at %d\n", 
	  inet_ntoa(addr.sin_addr), ntohs(addr.sin_port), 
	  strerror(errno), res, sock, (int) timeex(NULL));
    return;
  }
}

/*-----------------------------------------------------------------*/
static void
Heartbeat_HandleAck(int whichServ, CoDNSHeartbeatInfo *hb)
{
  PeerInfo *pi;
  struct timeval currentTime;
  int diff;

  if (whichServ < 0)
    return;

  pi = &g_allPeers[whichServ];

  /* see if this guy is in time */
  if (pi->pi_hbLastSentTime == hb->hi_ID) 
    pi->pi_hbGoodAckHistory |= 1;
  
  /* caculate RTT in MSec */
  gettimeofdayex(&currentTime, NULL);
  diff = (int)DiffTimesToMSec(currentTime, hb->hi_sendTime);
  if (diff < 0) {
    TRACE_L("weirdAck: %s took %d ms\n", pi->pi_name, diff);
    return;
  }

  /* calculate the average ping response time */
  pi->pi_avgPingTime = DecayAvg(pi->pi_pingCount == 0 ? 0 : pi->pi_avgPingTime,
				diff, &pi->pi_pingCount, VALID_PING_COUNT);

  /* average local DNS lookup time of the peer in MSecs */
  pi->pi_avgLocalLookupMS = hb->hi_avgLocalLookupMS;

  /* update the name server */
  pi->pi_nameServers[0] = hb->hi_nameServers[0];
  pi->pi_nameServers[1] = hb->hi_nameServers[1];

  /* update the neighbor node list 
     with the new response time */
  UpdateNeighborsList(pi);
}
/*-----------------------------------------------------------------*/
void
Heartbeat_Receive(int fd, void *token)
{
  /* get the heartbeat, send back an ack, and update our activity
     counter */
  struct sockaddr_in foreign_sin;
  int foreign_len = sizeof(foreign_sin);
  int res;
  CoDNSHeartbeatInfo hb;
  int whichServ = -1;
  int i;
  int portNum;

  /* compare with my name server to see if it's the same one */
#define IS_MY_NAME_SERVER(x) ((x.s_addr) != 0 &&                           \
			      ((x.s_addr) == g_MyNameServers[0].s_addr ||  \
			       (x.s_addr) == g_MyNameServers[1].s_addr))

  memset(&foreign_sin, 0, sizeof(foreign_sin));
  memset(&hb, 0, sizeof(hb));

  res = recvfrom(fd, &hb, sizeof(hb), 0, 
		 (struct sockaddr *) &foreign_sin, (socklen_t *)&foreign_len);
  if (res < 0) {
    if (errno == EAGAIN) {
      TRACE_L("Data doesn't seem to be available now\n");
      return;
    }
    TRACE_L("Error in recv from %s:%d for %s res: %d at socket %d at %d\n", 
	  inet_ntoa(foreign_sin.sin_addr), ntohs(foreign_sin.sin_port), 
	  strerror(errno), res, fd, (int) timeex(NULL));
    TRACE_L( "addr_len is %d\n", foreign_len);
    return;
  }
  ntoh_CoDNSHeartbeatInfo(&hb);

  /* Limin says some NAT boxes change port numbers on UDP, so if you
     send an ack back to the incoming port number, you're fine */
  portNum = ntohs(foreign_sin.sin_port);

  /* find which server sent us the heartbeat */
  for (i = 0; i < g_numPeers; i++) {
    if (foreign_sin.sin_addr.s_addr == g_allPeers[i].pi_address.s_addr) {
      whichServ = i;
      break;
    }
  }

  /* if we got an ack, figure out what it means and record it */
  if (hb.hi_statusVec & CODNS_HBI_ACK) {
    Heartbeat_HandleAck(whichServ, &hb);
    return;
  }

  if (hb.hi_statusVec != CODNS_HBI_QUERY) {
    /* we didn't understand what was being asked */
    return;
  }

  /* we ignore the peers sharing the same set of name servers */
  if (IS_MY_NAME_SERVER(hb.hi_nameServers[0]) 
      || IS_MY_NAME_SERVER(hb.hi_nameServers[1]))
    return;

  /* send an ack */
  Heartbeat_Send(fd, foreign_sin.sin_addr, portNum, 
		 CODNS_HBI_ACK, hb.hi_ID, &hb.hi_sendTime);
}

/*-----------------------------------------------------------------*/
static int
HeartBeatSock_Prepare(void) 
{
  /* socket for listening and sending heartbeat messages */
  struct sockaddr_in hb_sin;
  int sock;

  if ((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    TRACE_L1("Error in allocate heart beat socket\n");
    NiceExit(2, "Error in allocate heart beat socket");
  } 

  /* bind it */
  memset(&hb_sin, 0, sizeof(hb_sin));
  hb_sin.sin_family = AF_INET;
  hb_sin.sin_addr.s_addr = INADDR_ANY;
  hb_sin.sin_port = htons(CODNS_HB_PORT);
  if (bind(sock, (struct sockaddr *) &hb_sin, sizeof(hb_sin)) < 0) {
    TRACE_L("Error in bind heartbeat socket %d at port %d \n", 
	    sock, CODNS_HB_PORT);
    NiceExit(2, "Error in bind heartbeat socket");
  }
  
  /* make listen socket nonblocking */
  if (fcntl(sock, F_SETFL, O_NDELAY) == -1) {
    close(sock);
    NiceExit(2, "making listen socket nonblocking");
  }

  return(sock);
}
/*-----------------------------------------------------------------*/
static void
PrintNeighbors(int* neighbors, int num)
{
  int i;
  PeerInfo* pi;
  time_t now = timeex(NULL);

  TRACE_L("%d %s", (int) now, ctime(&now));
  for (i = 0; i < num; i++) {
    pi = &g_allPeers[neighbors[i]];
    TRACE_L("%dth : %s, ping(%d ms), DNS(%d ms)\n",
	  i, pi->pi_name, pi->pi_avgPingTime, pi->pi_avgLocalLookupMS);
  }
}
/*-----------------------------------------------------------------*/
static void
SendOff(int sock, int idx) 
{
  PeerInfo* pi;

  /* check if idx is in range */
  if (idx < 0 || idx >= MAX_PEERS)
    return;

  pi = &g_allPeers[idx];          

  /* skip when we don't know the address */
  if (pi->pi_address.s_addr == 0) 
    return;
  
  pi->pi_hbLastSentTime = timeex(NULL);
  Heartbeat_Send(sock, pi->pi_address, CODNS_HB_PORT,
		 CODNS_HBI_QUERY, pi->pi_hbLastSentTime, NULL);
  pi->pi_hbGoodAckHistory <<= 1;
}
/*-----------------------------------------------------------------*/
inline int
IsNeighbor(const int* neighbors, int num, int idx)
{
  /* see if idx is my neighbor node */
  int i;

  for (i = 0; i < num; i++)
    if (neighbors[i] == idx)
      return 1;

  return 0;
}
/*-----------------------------------------------------------------*/
void
SendHeartbeatQueries(int sock)
{
  /* a. sends a hearbeat to my neighbors
     b. keeps searching for nearest neighbors */
  static int neighbors[MAX_NEIGHBORS];
  static int numNeighbors = 0; /* current number of my neighbors */
  static int idx_p = -1;       /* next peer idx for HB */
  static int idx_n = -1;       /* next neighbor idx for HB */

  /* we've come around again? */
  if (idx_p == 0) {
    /* for debugging, print out our neighbors */
    PrintNeighbors(neighbors, numNeighbors);

    /* update my neighbors */
    GetNeighborsSnapshot(neighbors, &numNeighbors, FALSE);
    idx_n = -1; /* because 1 will be added to it soon */
  }

  /* send a hearbeat to a node which is not my neighbor */
  do {
    idx_p = (idx_p + 1) % g_numPeers;
    if (idx_p == 0)
      break;
  } while (IsNeighbor(neighbors, numNeighbors, idx_p));
  if (idx_p != 0 || !IsNeighbor(neighbors, numNeighbors, idx_p))
    SendOff(sock, idx_p);

  /* monitor my neighbor's health */
  if (numNeighbors > 0 && idx_n < numNeighbors) {
    idx_n = (idx_n + 1) % numNeighbors;
    SendOff(sock, neighbors[idx_n]);
  }
}
/*-----------------------------------------------------------------*/
static void
ReadResolvConf(void)
{
  /* read in first two name servers in /etc/resolv.conf and convert them
     into numeric forms for heartbeat information */

#define MAX_LINE 1024
  FILE* fp;
  const char* file = "/etc/resolv.conf";
  char line[MAX_LINE], *s, *p;
  struct in_addr addr;
  int count = 0;

  /* initialize the array with 0 */
  memset(g_MyNameServers, 0, sizeof(g_MyNameServers));

  /* file readable ? */
  if (access(file, R_OK) != 0)
    return;

  /* open and read the contents */
  if ((fp = fopen(file, "r")) == NULL)
    return;

#define NAMESERVER "nameserver"
  while (fgets(line, MAX_LINE, fp)) {
    if ((s = strstr(line, NAMESERVER)) != NULL) {
      s += sizeof(NAMESERVER);
      p = line;

      /* search IP address */
      while (*s && !(isdigit((int)*s) || *s == '.'))
	s++;

      /* copy IP address */
      while (*s && (isdigit((int)*s) || *s == '.'))
	*p++ = *s++;
      *p = 0;
      
      /* convert into numeric forms */
      if (inet_aton(line, &addr) != 0) 
	g_MyNameServers[count++] = addr;

      /* read in only the first two */
      if (count >= 2) 
	break;
    }
  }
  fclose(fp);
}

/*-----------------------------------------------------------------*/
int 
Heartbeat_Init()
{
  ReadResolvConf();
  return HeartBeatSock_Prepare();
}
