#include <sys/types.h>
#include <sys/socket.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include "applib.h"
#include "codns_int.h"
#include "codnsstat.h"
#include "debug.h"
#include "codnslog.h"

#define LOCAL_ALL         0
#define LOCAL_GOOD        1
#define LOCAL_FAIL        2
#define REMOTE_ALL        3
#define REMOTE_GOOD       4
#define REMOTE_FAIL       5
#define REMOTE_SQ         6
#define OVERALL_SUCCESS   7
#define MAX_AVERAGES      8

/* averages & counts */
static double g_allAverages[MAX_AVERAGES];
static int g_allCounts[MAX_AVERAGES];

extern HANDLE hdebugLog2;
/*-----------------------------------------------------------------*/
static void
RecordResponse(QuerySource* query, int msElapsed, 
	       int isSuccessful, int isLocal)
{
  char resp[2048];
  int size;
  struct in_addr* p_addr;
  int i;

  /* if the query is for internal use, forget it */
  if (query->qs_queryinfo.lqi_id == INT_QUERY_ID)
    return;

  p_addr = query->qs_query.nsq_result.nsq_address;
  size = snprintf(resp, sizeof(resp),
		  "[DNSLog5] %09d.%03d %s %c %c %d %d %c %d ",
		  (int)query->qs_startTime.tv_sec, /* starting time(sec) */
		  (int)(query->qs_startTime.tv_usec /1000),
		  query->qs_query.nsq_name,    /* name */
		  isLocal ?      'L' : 'R',    /* local or remote */
		  isSuccessful ? 'G' : 'B',    /* good or bad(fail) */
		  msElapsed,                   /* response time */
		  isLocal ? 
		  -1 : query->qs_query.nsq_delay,    /* delay of r-query */
		  p_addr[0].s_addr != 0 ? '+' : '-', /* valid address or not */
		  (query->qs_queryinfo.lqi_id == HBTWGET_CODNS_ID) ? 
		  query->qs_queryinfo.lqi_cache : 0); 
  

  /* let's write down the IP addresses */
  for (i = 0; i < MAX_ANSWERS && p_addr[i].s_addr != 0; i++) {
    struct in_addr temp = p_addr[i];
    int s;

    s = snprintf(resp+size, sizeof(resp)-size-2, "%s ", inet_ntoa(temp));
    if (s < sizeof(resp) - size - 2) {
      size += s;
    } else {
      /* we're out of memory, so remove the last IP address */
      break; 
    }
  }

  if (sizeof(resp) > size-2) {
    snprintf(resp+size, sizeof(resp)-size, "\n");
    size++;
  }
  resp[sizeof(resp)-1] = 0; /* for safety */
  write(g_logFd, resp, size);
}

/*-----------------------------------------------------------------*/
static double
GetStddevPeerRespTime()
{
  double var, stddev;

  var = g_allAverages[REMOTE_SQ] - 
       (g_allAverages[REMOTE_GOOD] * g_allAverages[REMOTE_GOOD]);
  if (var < 0) stddev = 0;
  else stddev = sqrt(var); 

  return stddev;
}

/*-----------------------------------------------------------------*/
void
RecordRemoteQuery(QuerySource* query, int serv)
{
  /* record a log for remote requests at the time of remote querying */
  char resp[2048];
  int size;

  /* if the query is for internal use, forget it */
  if (query->qs_queryinfo.lqi_id == INT_QUERY_ID)
    return;

  /* the log is similar to the response logs except for "R S" in the middle,
     R: remote, S : sending, * : not yet known  */
  size = snprintf(resp, sizeof(resp),
	   "[DNSLog5] %09d.%03d %s R S %d %d * %s %d\n",
	   (int)query->qs_startTime.tv_sec, /* starting time(sec) */
	   (int)(query->qs_startTime.tv_usec /1000), /* starting time(msec) */
	   query->qs_query.nsq_name,        /* name */
	   query->qs_numUDPs,               /* how many UDPs ? */
	   query->qs_query.nsq_delay,       /* delay from the request arrival*/
	   g_allPeers[serv].pi_name,        /* destination server name */
	   (query->qs_queryinfo.lqi_id == HBTWGET_CODNS_ID) ? 
	   query->qs_queryinfo.lqi_cache : 0);       

  resp[sizeof(resp)-1] = 0;
  write(g_logFd, resp, size);
}

/*-----------------------------------------------------------------*/
void
UpdateDNSStatistics(QuerySource* query, int isLocal)
{
  /* each time name lookup is finished, this function is called.
     here is a good place to calculate the statistics 
  */
  int msElapsed, idx, isSuccessful;
  struct timeval currTime;
  static int init = 1;

  /* current time */
  gettimeofdayex(&currTime, NULL);

  /* initialize ? */
  if (init == 1) {
    memset(g_allAverages, 0, sizeof(g_allAverages));
    memset(g_allCounts, 0, sizeof(g_allCounts));
    init = 0;
  }
  
  /* is this faster response */
  isSuccessful = (query->qs_fd >= 0);

  /* measure response time */
  msElapsed = (int)DiffTimesToMSec(currTime, query->qs_startTime);

  /* leave a log */
  RecordResponse(query, msElapsed, isSuccessful, isLocal);

  /* specific statistics */
  if (isLocal)
    idx = (isSuccessful) ? LOCAL_GOOD : LOCAL_FAIL;
  else 
    idx = (isSuccessful) ? REMOTE_GOOD : REMOTE_FAIL;

  g_allAverages[idx] = DecayAvg(g_allAverages[idx], msElapsed,
                                &g_allCounts[idx], 100);

  if (idx == REMOTE_GOOD) {
    /* calcuate E(x^2) for the case of peer lookup being fast 
       this value is used for stddev calculation */
    idx = REMOTE_SQ;
    g_allAverages[idx] = DecayAvg(g_allAverages[idx], 
				  msElapsed*msElapsed,
				  &g_allCounts[idx], 100);
  }

  /* all statistics */
  idx =  (isLocal) ? LOCAL_ALL : REMOTE_ALL;
  g_allAverages[idx] = DecayAvg(g_allAverages[idx], msElapsed,
				&g_allCounts[idx], 100);

  /* overall average response time of successful ones */
  if (isSuccessful) {
    idx = OVERALL_SUCCESS;
    g_allAverages[idx] = DecayAvg(g_allAverages[idx], msElapsed,
				  &g_allCounts[idx], 100);
  }

  /* update average local DNS lookup time */
  if (isLocal)  
    g_avgLocalLookupMS = g_allAverages[LOCAL_ALL];
}

/*-----------------------------------------------------------------*/
void
CheckStatQuery(const SelCtx* ctx)
     //int statSock, fd_set* ptempSet)
{
  /* broadcast the statistics outside */
  char buf[32*1024]; 
  int newFD, numBytes;

  if (!FD_ISSET(ctx->sel_fd, ctx->sel_tempSet)) 
    return;

  if ((newFD = accept(ctx->sel_fd, NULL, NULL)) < 0)
    return;

  /* make it non-blocking */
  if (fcntl(newFD, F_SETFL, O_NDELAY) < 0) {
    TRACE_L1("fcntl failed\n");
    close(newFD);
    return;
  }

  /* wrap the statistics up */
  numBytes = snprintf(buf, sizeof(buf),
		      "%s, Average(ms) : %.1f %.1f all, %.1f %.1f gd, %.1f %.1f fl %.1f suc %.1f p-stddev\n",
		      g_localHostname,
		      g_allAverages[LOCAL_ALL],
		      g_allAverages[REMOTE_ALL],
		      g_allAverages[LOCAL_GOOD],
		      g_allAverages[REMOTE_GOOD],
		      g_allAverages[LOCAL_FAIL],
		      g_allAverages[REMOTE_FAIL],
		      g_allAverages[OVERALL_SUCCESS],
		      GetStddevPeerRespTime());

  /* send and close */
  write (newFD, buf, numBytes);
  close(newFD);
}
/*-----------------------------------------------------------------*/
int
GetUDPStartDelay(int udpWinsHistory, int base, int max)
{
  /* estimate the starting delay in microseconds
     UDPwinsHistory, successful UDP average time */

  int delay, udpWinCount32, udpWinCount16;

  delay = base;               /* base is used when local DNS is 100% better */
  udpWinCount32 = PopCount(udpWinsHistory & 0xffffffff);
  udpWinCount16 = PopCount(udpWinsHistory & 0xffff);

  /* adjust according to udp query wins history */
  if (udpWinCount32 <= 3) {         /* local DNS is more than 90% better */
    if (udpWinCount32 != 0) {        
      if (udpWinCount16 == 0)       /* 100 % for recent 16 lookups */
	delay = (int)(base * 0.7);
      else                          /* showing some problems for recent 16 */
	delay = (int)(base * 0.4);
    }
  } else if (udpWinCount16 >= 8) {  /* local DNS is in bad shape */
    delay = 0;                      /* send a UDP query right away */
  }

  return delay;
}

/*-----------------------------------------------------------------*/
double 
GetMostPeerRespTime(void)
{
  double stddev = GetStddevPeerRespTime();

  /* 99 %-tile peer lookup response time or 150 ms 
     we assume Gauss distribution, but it may not be the case ^^ */
  return __min(g_allAverages[REMOTE_GOOD] + 3*stddev, 150.0);
}
/*-----------------------------------------------------------------*/
