#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/uio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#ifdef OS_DARWIN
#include <nameser.h>
#endif
#include <resolv.h>
#include "queue.h"
#include "applib.h"
#include "codns_int.h"
#include "codnshbt.h"
#include "HRW_codns.h"
#include "codnsstat.h"
#include "debug.h"
#include "codnslog.h"
#include "ports.h"

#define CODNS_VERS 5
#define COHOST_RATELIMIT 20

/* fds for logging processes */
int g_logFd;                      /* basic debug trace */
HANDLE hdebugLog;		  /* for debugging log */

/* Time interval for UDP request throttling */
unsigned long g_udelay = 1000000 / COHOST_RATELIMIT;

int g_opt_timeout = 60;           /* when we stop waiting for responses, in seconds */
int g_opt_verbose = 0;
int g_opt_ratelimit = COHOST_RATELIMIT;         /* Rate limit for UDP queries, never more than x per second */

/* Kind of like QuerySource, but simpler. */
typedef struct query_data {
  char *qd_name;
  int qd_token;
  struct in_addr qd_addr;
} QueryData;

void 
TimevalSub(const struct timeval *tv1,
	   const struct timeval *tv2,
	   struct timeval *tv_out)
{
  struct timeval tmp1, tmp2;
  struct timeval *t1 = &tmp1, *t2 = &tmp2;

  if (!tv1 || !tv2 || !tv_out)
    return;

  if (tv1->tv_sec < tv2->tv_sec ||
      (tv1->tv_sec == tv2->tv_sec && tv1->tv_usec < tv2->tv_usec))
    {
      tv_out->tv_sec  = 0;
      tv_out->tv_usec = 0;
      return;
    }

  tmp1 = *tv1;
  tmp2 = *tv2;

  tv_out->tv_sec = t1->tv_sec - t2->tv_sec;
  
  if (t1->tv_usec < t2->tv_usec)
    {
      tv_out->tv_sec -= 1;
      tv_out->tv_usec = t1->tv_usec + 1000000 - t2->tv_usec;      
    }
  else
    {
      tv_out->tv_usec  = t1->tv_usec  - t2->tv_usec;
    }
}


void Help()
{
  printf("USAGE: cohost-bin [-h] [-s <num>] [-t <secs>] [-v] <name> <servers>\n");
  printf("NOTE: this is not meant to be run directly, but through cohost.\n");
  printf("\n");
  printf("    -h         This help message.\n");
  printf("    -s <num>   Slow down requests to <num>/s.\n");
  printf("    -t <secs>  Timeout before returning (%is).\n", g_opt_timeout);
  printf("    -v         Verbose operation, printed to stderr.\n");
  exit(0);
}

/* So this builds without codns.c */
void InitLogFile()
{
  hdebugLog = CreateLogFHandle("dbgcodns", TRUE);

  if (hdebugLog == NULL || (OpenLogF(hdebugLog) < 0)) {
    printf("Couldn't open log file.\n");
    return;
  }
}


QueryData *InitPeers(int num_peers, char** names, int *num_resolved)
{
  QueryData *qds, *qd;
  struct hostent *hp = NULL;
  int i;

  qd = qds = calloc(num_peers, sizeof(QueryData));
  *num_resolved = 0;

  /* Find IP addresses for all CoDNS nodes given. */
  
  for (i = 0; i < num_peers; i++, qd++)
    {
      if ( (hp = gethostbyname(names[i])) != NULL)
	{
	  qd->qd_name = names[i];
	    
	  /* Pick first IP address returned. */
	  qd->qd_addr = *((struct in_addr *)(hp->h_addr_list[0]));      
	  qd->qd_token = random();
	  (*num_resolved)++;
	}
    }
  
  return qds;
}

int GetSocket(void)
{
  int sock;

  if ( (sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    return -1;
  
  return sock;
}

int SendRequests(int sock, char *name, int num_peers, QueryData *qds)
{
  NSQuery nsq;
  QueryData *qd = qds;
  struct sockaddr_in addr;  
  int i, num_sent = 0;
  
  /* Go over list of peers, fill the corresponding query data
   * into the name server query, and launch requests. Reports
   * back the number of requests sent.
   */
  for (i = 0; i < num_peers; i++, qd++)
    {
      if (qd->qd_name == NULL)
	continue;
      
      memset(&addr, 0, sizeof(addr));
      addr.sin_family = AF_INET;
      addr.sin_addr = qd->qd_addr;
      addr.sin_port = htons(CODNS_PORT);
      
      memset(&nsq, 0, sizeof(NSQuery));
      nsq.nsq_version = CODNS_VERS;
      strncpy(nsq.nsq_name, name, MAX_NSQUERY_NAME);
      nsq.nsq_name[MAX_NSQUERY_NAME-1] = 0;
      nsq.nsq_randomToken = qd->qd_token;
      nsq.nsq_id = qd->qd_token;      
      hton_NSQuery(&nsq);
      
      if (sendto(sock, &nsq, sizeof(NSQuery), 0,
		 (struct sockaddr *) &addr,
		 sizeof(struct sockaddr_in)) == sizeof(NSQuery))
	num_sent++;

      /* Wait for a little while to throttle requests */
      usleep(g_udelay);
    }
  
  return num_sent;
}

void PrintResponse(QueryData *qd, NSQuery *nsq)
{
  int i;
  printf("%s(%s): ", qd->qd_name, inet_ntoa(qd->qd_addr));

  for (i = 0; i < MAX_ANSWERS && nsq->nsq_result.nsq_address[i].s_addr != 0; i++)
    printf("%s ", inet_ntoa(nsq->nsq_result.nsq_address[i]));
  printf("\n");
}

int ReadResponses(int sock, int num_sent, int num_peers, QueryData *qds)
{
  NSQuery nsq;
  struct sockaddr_in from;
  int fromLen = sizeof(struct sockaddr_in);
  struct timeval tRemain = { g_opt_timeout, 0 };
  struct timeval t0, t1, tTmp;
  fd_set set;
  int i, num_read = 0;
  
  if (fcntl(sock, F_SETFL, O_NDELAY) == -1)
    {
      printf("failed to make the socket non-blocking\n");
      return 0;
    }
  
  FD_ZERO(&set);
  FD_SET(sock, &set);

  while (1)
    {
      gettimeofdayex(&t0, NULL);
      
      if (select(sock+1, &set, NULL, NULL, &tRemain) < 0)
	{
	  printf("select() error: %s\n", strerror(errno));
	  return num_read;
	}
      
      gettimeofdayex(&t1, NULL);
      
      if (FD_ISSET(sock, &set))
	{
	  int res = recvfrom(sock, &nsq, sizeof(NSQuery), 0,
			     (struct sockaddr *) &from,
			     (socklen_t *) &fromLen);
	  if (res <= 0)
	    return 0;
	  
	  ntoh_NSQuery(&nsq);
	  
	  /* Find the query corresponding to this response.
	   * Linear search, lame.
	   */
	  for (i = 0; i < num_peers; i++)
	    {
	      if (qds[i].qd_token == nsq.nsq_randomToken)
		{
		  PrintResponse(&qds[i], &nsq);
		  num_read++;	      
		  break;
		}
	    }
	  
	}
      else
	{
	  /* Nothing read, so we returned due to the
	   * timeout -- we can return.
	   */
	  break;
	}
      
      /* If we've received everything we asked for,
       * return.
       */
      if (num_read == num_sent)
	break;
      
      /* Update timeout interval and try again.
       */
      TimevalSub(&t1, &t0, &tTmp);
      TimevalSub(&tRemain, &tTmp, &tRemain);
    }

  return num_read;
}


int main(int argc, char** argv)
{
  QueryData *qds = NULL; /* QueryData, to associate queries/responses */
  char *name = NULL; /* The name to resolve */
  struct timeval tv;
  int c, sock, num_peers = 0, num_resolved = 0, num_sent = 0, num_rcvd = 0;

  if (argc <= 1)
    Help();
 
  while ( (c = getopt(argc, argv, "ht:s:v")) != -1)
    {
      switch (c) {
      case 'h':
	Help();
	break;
	
      case 's':
	g_opt_ratelimit = atoi(optarg);
	
	if (g_opt_ratelimit <= 0)
	  g_opt_ratelimit = COHOST_RATELIMIT;

	if (g_opt_ratelimit > 50)
	  g_opt_ratelimit = 50;
	break;

      case 't':
	g_opt_timeout = atoi(optarg);
	break;

      case 'v':
	g_opt_verbose = 1;
	break;
      }
    }

  /* initialize random seed */
  gettimeofdayex(&tv, NULL);
  srandom(tv.tv_usec);
  
  /* translate the request rate limit into a waiting time */
  g_udelay = 1000000 / g_opt_ratelimit;

  if (optind == argc)
    {
      printf("Please provide a name to look up plus CoDNS nodes to query.\n");
      return -1;
    }

  name = argv[optind++];
  
  if (optind == argc)
    {
      printf("I need some CoDNS nodes to contact.\n");
      return -2;
    }

  num_peers = argc - optind;
  qds = InitPeers(num_peers, argv + optind, &num_resolved);

  if (g_opt_verbose)
    fprintf(stderr, "%i/%i peers initialized.\n", num_resolved, num_peers);
  
  /* Abort if we got nothing. */
  if (num_resolved == 0)
    return 0;
    
  if ( (sock = GetSocket()) < 0)
    {
      printf("Could not obtain socket.\n");
      return -3;
    }
  
  num_sent = SendRequests(sock, name, num_peers, qds);
  if (g_opt_verbose)
    fprintf(stderr, "%i/%i requests sent.\n", num_sent, num_peers);
  
  num_rcvd = ReadResponses(sock, num_sent, num_peers, qds);
  if (g_opt_verbose)
    fprintf(stderr, "%i/%i responses received.\n", num_rcvd, num_peers);
  
  close(sock);
  
  return 0;
}
