#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include "applib.h"
#include "codns_int.h"
#include "codnslog.h"

/* all defined in codns.c */
extern int IsConfigFileUpdated(char* filename);
extern void ReadConfigFile(int readOptionsOnly);

extern char* g_autoUpdateHost;
extern char* g_autoUpdateFile;

/*-----------------------------------------------------------------*/
static void
MakeRequest(char *buf, int size)
{
  snprintf(buf, size,
	   "GET http://%s%s HTTP/1.0 \r\n"
	   "Host: %s\r\n\r\n",
	   g_autoUpdateHost, g_autoUpdateFile, g_autoUpdateHost);
}
/*-----------------------------------------------------------------*/
static int
GetHeaderBlock(int s, char* buf, int maxLen, int* bodyOffset, int* contLen) 
{
  /* receive the header block of a HTTP response
     returns the response code. -1 means an error */
  int used = 0;
  int res;
  char* temp;
  int rcode = -1;

  *bodyOffset = *contLen = -1;
  while (1) {
    res = read(s, buf+used, maxLen-used);
    if (res <= 0)
      return -1;
    
    used += res;
    if (used >= maxLen)
      return -1;

    buf[used] = 0;
    /* see if we've seen the end of a header */
    if ((temp = strstr(buf, "\n\r\n")) != NULL) {
      *bodyOffset = temp + 3 - buf;
      break;
    } else if ((temp = strstr(buf, "\n\n")) != NULL) {
      *bodyOffset = temp + 2 - buf;
      break;
    }
  }

  if ((temp = GetField(buf, 1)) == NULL) {
    TRACE_L("failed in getting the response code buf=%s", buf);
    return -1;
  }
  rcode = atoi(temp);

  /* find the content length */
  if ((temp = strncasestr(buf, used, "/ncontent-length:")) != NULL) {
    temp += sizeof("/ncontent-length:")-1;
    while (*temp && isspace((int)*temp)) temp++;
    *contLen = atoi(temp);
  }

  return rcode;
}
/*-----------------------------------------------------------------*/
static void 
ConfigFileAutoUpdateMain(void)
{
  /* periodically download the new config file(codns.conf)
     if it's new then rename it into codns.conf.new so that
     the other processes can adapt to it */
  int s;
  int res;
  int rcode;
  int bodyOffset;
  int bodyWritten;
  int len;
  int contLen;
  char request[2048];
  char buf[32*1024];
  char tempfile[MAX_PATH];
  struct sockaddr_in saddr;
  struct timeval now;

#define RAND_DIVISOR 3607  /* first prime number > 3600 */
#define HTTP_PORT 80

  /* in order not to flood the server with the same requests 
     from everywhere, we wait a random amount of time before downloading */
  gettimeofdayex(&now, NULL);
  srand(now.tv_usec);

  while (1) {
    int fd = -1; /* tempfile fd */
    int randsecs = rand() % RAND_DIVISOR;
    struct in_addr addr;

    /* set the tempfile name to NULL */
    tempfile[0] = 0;

    /* sleep at least 10 secs */
    sleep(10 + randsecs);

    /* do a name lookup of the server */
    if (g_autoUpdateHost == NULL || 
	CoDNSGetHostByNameSync(g_autoUpdateHost, &addr) < 0) {
      if (g_autoUpdateHost == NULL) {
	TRACE_L("autoUpdateHost = NULL");
      } else {
	TRACE_L("CoDNS lookup failed\n");
      }
      continue;
    }

    /* create a socket */
    if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      TRACE_L("socket creation failure");
      exit(-1);
    }

    /* connect to the server */
    saddr.sin_family = AF_INET;
    saddr.sin_addr = addr;
    saddr.sin_port = htons(HTTP_PORT);
    if (connect(s, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) 
      goto error;

    /* make an HTTP request */
    MakeRequest(request, sizeof(request));
      
    /* send the request */
    if (write(s, request, strlen(request)) != strlen(request)) {
      TRACE_L("writing request failed\n");
      goto error;
    }

    /* read the header block */
    rcode = GetHeaderBlock(s, buf, sizeof(buf), &bodyOffset, &contLen);
    if (rcode != 200) {
      TRACE_L("response header not 200 buf=%s", buf);
      goto error;
    }
    
    /* make a tempfile for writing into */
    snprintf(tempfile, sizeof(tempfile), "codns.conf.XXXXXX");
    if ((fd = mkstemp(tempfile)) < 0) {
      TRACE_L("making tempfile failed\n");
      goto error;
    }
    
    /* write body into the tempfile */
    len = strlen(&buf[bodyOffset]);
    bodyWritten = write(fd, &buf[bodyOffset], len);
    if (bodyWritten != len) {
      TRACE_L("write error len=%d\n", len);
      goto error;
    }

    while (contLen == -1 || bodyWritten < contLen) {
      res = read(s, buf, sizeof(buf));
      if (res == 0 && contLen == -1) 
	break;
      if (res <= 0) 
	goto error;

      len = write(fd, buf, res);
      if (len < res) {
	TRACE_L("fwrite error written=%d\n", len);
	goto error;
      }
      bodyWritten += len;
    }

    /* close the temp file */
    close(fd);    
    fd = -1;

    /* see if we need to update this file */
    if (IsConfigFileUpdated(tempfile)) {
      /* it's sufficient to read only the options part */
      if (rename(tempfile, NEWCONFIGFILE) == 0)
	ReadConfigFile(TRUE); 
    }

  error:
    close(s);
    if (fd >= 0) close(fd);
    if (*tempfile && access(tempfile, F_OK) == 0) 
      unlink(tempfile);
    
    /* sleep at least for 10 hours */
    randsecs = rand() % RAND_DIVISOR;
    sleep(36000 + randsecs);
  }
}
/*-----------------------------------------------------------------*/
void
CreateConfigFileAutoUpdater(void)
{
  if (fork() > 0) 
    return;

  /* i'm in a child */
  ConfigFileAutoUpdateMain();
  TRACE_L("exit from the child\n");
}
